# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `gumbel_cdf.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.bijectors import gumbel_cdf
from distrax._src.bijectors import tanh
import jax
import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp

tfb = tfp.bijectors

RTOL = 1e-5


class GumbelCDFTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.seed = jax.random.PRNGKey(1234)

  def test_properties(self):
    bijector = gumbel_cdf.GumbelCDF()
    self.assertEqual(bijector.event_ndims_in, 0)
    self.assertEqual(bijector.event_ndims_out, 0)
    self.assertFalse(bijector.is_constant_jacobian)
    self.assertFalse(bijector.is_constant_log_det)

  @chex.all_variants
  @parameterized.parameters(
      {'x_shape': (2,)},
      {'x_shape': (2, 3)},
      {'x_shape': (2, 3, 4)})
  def test_forward_shapes(self, x_shape):
    x = jnp.zeros(x_shape)
    bijector = gumbel_cdf.GumbelCDF()
    y1 = self.variant(bijector.forward)(x)
    logdet1 = self.variant(bijector.forward_log_det_jacobian)(x)
    y2, logdet2 = self.variant(bijector.forward_and_log_det)(x)
    self.assertEqual(y1.shape, x_shape)
    self.assertEqual(y2.shape, x_shape)
    self.assertEqual(logdet1.shape, x_shape)
    self.assertEqual(logdet2.shape, x_shape)

  @chex.all_variants
  @parameterized.parameters(
      {'y_shape': (2,)},
      {'y_shape': (2, 3)},
      {'y_shape': (2, 3, 4)})
  def test_inverse_shapes(self, y_shape):
    y = jnp.zeros(y_shape)
    bijector = gumbel_cdf.GumbelCDF()
    x1 = self.variant(bijector.inverse)(y)
    logdet1 = self.variant(bijector.inverse_log_det_jacobian)(y)
    x2, logdet2 = self.variant(bijector.inverse_and_log_det)(y)
    self.assertEqual(x1.shape, y_shape)
    self.assertEqual(x2.shape, y_shape)
    self.assertEqual(logdet1.shape, y_shape)
    self.assertEqual(logdet2.shape, y_shape)

  @chex.all_variants
  def test_forward(self):
    x = jax.random.normal(self.seed, (100,))
    bijector = gumbel_cdf.GumbelCDF()
    y = self.variant(bijector.forward)(x)
    expected_y = jnp.exp(-jnp.exp(-x))
    np.testing.assert_allclose(y, expected_y, rtol=RTOL)

  @chex.all_variants
  def test_forward_log_det_jacobian(self):
    x = jax.random.normal(self.seed, (100,))
    bijector = gumbel_cdf.GumbelCDF()
    fwd_logdet = self.variant(bijector.forward_log_det_jacobian)(x)
    actual = jnp.log(jax.vmap(jax.grad(bijector.forward))(x))
    np.testing.assert_allclose(fwd_logdet, actual, rtol=1e-3)

  @chex.all_variants
  def test_forward_and_log_det(self):
    x = jax.random.normal(self.seed, (100,))
    bijector = gumbel_cdf.GumbelCDF()
    y1 = self.variant(bijector.forward)(x)
    logdet1 = self.variant(bijector.forward_log_det_jacobian)(x)
    y2, logdet2 = self.variant(bijector.forward_and_log_det)(x)
    np.testing.assert_allclose(y1, y2, rtol=RTOL)
    np.testing.assert_allclose(logdet1, logdet2, rtol=RTOL)

  @chex.all_variants
  def test_inverse(self):
    x = jax.random.normal(self.seed, (100,))
    bijector = gumbel_cdf.GumbelCDF()
    y = self.variant(bijector.forward)(x)
    x_rec = self.variant(bijector.inverse)(y)
    np.testing.assert_allclose(x_rec, x, rtol=1e-3)

  @chex.all_variants
  def test_inverse_log_det_jacobian(self):
    x = jax.random.normal(self.seed, (100,))
    bijector = gumbel_cdf.GumbelCDF()
    y = self.variant(bijector.forward)(x)
    fwd_logdet = self.variant(bijector.forward_log_det_jacobian)(x)
    inv_logdet = self.variant(bijector.inverse_log_det_jacobian)(y)
    np.testing.assert_allclose(inv_logdet, -fwd_logdet, rtol=1e-3)

  @chex.all_variants
  def test_inverse_and_log_det(self):
    y = jax.random.uniform(self.seed, (100,))
    bijector = gumbel_cdf.GumbelCDF()
    x1 = self.variant(bijector.inverse)(y)
    logdet1 = self.variant(bijector.inverse_log_det_jacobian)(y)
    x2, logdet2 = self.variant(bijector.inverse_and_log_det)(y)
    np.testing.assert_allclose(x1, x2, rtol=RTOL)
    np.testing.assert_allclose(logdet1, logdet2, rtol=RTOL)

  @chex.all_variants
  def test_stability(self):
    bijector = gumbel_cdf.GumbelCDF()
    tfp_bijector = tfb.GumbelCDF()

    x = np.array([-10.0, -3.3, 0.0, 3.3, 10.0], dtype=np.float32)
    fldj = tfp_bijector.forward_log_det_jacobian(x, event_ndims=0)
    fldj_ = self.variant(bijector.forward_log_det_jacobian)(x)
    np.testing.assert_allclose(fldj_, fldj, rtol=RTOL)

    y = bijector.forward(x)  # pytype: disable=wrong-arg-types  # jax-ndarray
    ildj = tfp_bijector.inverse_log_det_jacobian(y, event_ndims=0)
    ildj_ = self.variant(bijector.inverse_log_det_jacobian)(y)
    np.testing.assert_allclose(ildj_, ildj, rtol=RTOL)

  def test_jittable(self):
    @jax.jit
    def f(x, b):
      return b.forward(x)

    bijector = gumbel_cdf.GumbelCDF()
    x = np.zeros(())
    f(x, bijector)

  def test_same_as(self):
    bijector = gumbel_cdf.GumbelCDF()
    self.assertTrue(bijector.same_as(bijector))
    self.assertTrue(bijector.same_as(gumbel_cdf.GumbelCDF()))
    self.assertFalse(bijector.same_as(tanh.Tanh()))


if __name__ == '__main__':
  jax.config.update('jax_threefry_partitionable', False)
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `mvn_tri.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions.mvn_tri import MultivariateNormalTri
from distrax._src.utils import equivalence
import jax.experimental
import jax.numpy as jnp
import numpy as np


def _get_scale_tril_from_scale_triu(scale_triu: np.ndarray) -> np.ndarray:
  scale_triu = np.triu(scale_triu)
  scale_triu_t = np.vectorize(np.transpose, signature='(k,k)->(k,k)')(
      scale_triu)
  cov = np.matmul(scale_triu, scale_triu_t)
  return np.linalg.cholesky(cov)


class MultivariateNormalTriTest(equivalence.EquivalenceTest):

  def setUp(self):
    super().setUp()
    self._init_distr_cls(MultivariateNormalTri)

  @parameterized.named_parameters(
      ('all inputs are None', {}),
      ('wrong dimension of loc', {
          'loc': np.array(0.),
      }),
      ('scale_tri is 0d', {
          'scale_tri': np.array(1.),
      }),
      ('scale_tri is 1d', {
          'scale_tri': np.ones((4,)),
      }),
      ('scale_tri is not square', {
          'scale_tri': np.ones((4, 3)),
      }),
      ('inconsistent loc and scale_tri', {
          'loc': np.zeros((4,)),
          'scale_tri': np.ones((5, 5)),
      }),
  )
  def test_raises_on_wrong_inputs(self, dist_kwargs):
    with self.assertRaises(ValueError):
      self.distrax_cls(**dist_kwargs)

  @parameterized.named_parameters(
      ('loc provided', {'loc': np.zeros((4,))}),
      ('scale_tri provided', {'scale_tri': np.eye(4)}),
  )
  def test_default_properties(self, dist_kwargs):
    dist = self.distrax_cls(**dist_kwargs)
    self.assertTrue(dist.is_lower)
    self.assertion_fn(rtol=1e-3)(dist.loc, jnp.zeros((4,)))
    self.assertion_fn(rtol=1e-3)(dist.scale_tri, jnp.eye(4))

  @parameterized.named_parameters(
      ('unbatched', (), (4,), (4, 4), True),
      ('batched loc', (7,), (7, 4), (4, 4), True),
      ('batched scale_tri lower', (7,), (4,), (7, 4, 4), True),
      ('batched scale_tri upper', (7,), (4,), (7, 4, 4), False),
  )
  def test_properties(self, batch_shape, loc_shape, scale_tri_shape, is_lower):
    rng = np.random.default_rng(2022)
    loc = rng.normal(size=loc_shape)
    scale_tri = rng.normal(size=scale_tri_shape)
    dist = self.distrax_cls(loc=loc, scale_tri=scale_tri, is_lower=is_lower)
    tri_fn = jnp.tril if is_lower else jnp.triu
    self.assertEqual(dist.batch_shape, batch_shape)
    self.assertEqual(dist.is_lower, is_lower)
    self.assertion_fn(rtol=1e-3)(
        dist.loc, jnp.broadcast_to(loc, batch_shape + (4,)))
    self.assertion_fn(rtol=1e-3)(dist.scale_tri, jnp.broadcast_to(
        tri_fn(scale_tri), batch_shape + (4, 4)))

  @chex.all_variants
  @parameterized.named_parameters(
      ('unbatched, no shape', (), (4,), (4, 4)),
      ('batched loc, no shape', (), (7, 4), (4, 4)),
      ('batched scale_tri, no shape', (), (4,), (7, 4, 4)),
      ('unbatched, with shape', (3,), (4,), (4, 4)),
      ('batched loc, with shape', (3,), (7, 4), (4, 4)),
      ('batched scale_tri, with shape', (3,), (4,), (7, 4, 4)),
  )
  def test_sample_shape(self, sample_shape, loc_shape, scale_tri_shape):
    rng = np.random.default_rng(2022)
    loc = rng.normal(size=loc_shape)
    scale_tri = rng.normal(size=scale_tri_shape)
    dist_kwargs = {'loc': loc, 'scale_tri': scale_tri}
    tfp_dist_kwargs = {'loc': loc, 'scale_tril': scale_tri}
    super()._test_sample_shape(
        dist_args=(), dist_kwargs=dist_kwargs, tfp_dist_kwargs=tfp_dist_kwargs,
        sample_shape=sample_shape)

  @chex.all_variants
  @parameterized.named_parameters(
      ('float32', jnp.float32),
      ('float64', jnp.float64))
  def test_sample_dtype(self, dtype):
    with jax.experimental.enable_x64(dtype.dtype.itemsize == 8):
      dist_params = {
          'loc': np.array([0., 0.], dtype),
          'scale_tri': np.array([[1., 0.], [0., 1.]], dtype)}
      dist = self.distrax_cls(**dist_params)
      samples = self.variant(dist.sample)(seed=self.key)
      self.assertEqual(samples.dtype, dist.dtype)
      chex.assert_type(samples, dtype)

  @chex.all_variants
  @parameterized.named_parameters(
      ('unbatched, unbatched value', (4,), (4,), (4, 4), True),
      ('unbatched, unbatched value, upper', (4,), (4,), (4, 4), False),
      ('batched loc, unbatched value', (4,), (7, 4), (4, 4), True),
      ('batched scale_tri, unbatched value', (4,), (4,), (7, 4, 4), True),
      ('unbatched, batched value', (3, 7, 4), (4,), (4, 4), True),
      ('batched loc, batched value', (3, 7, 4), (7, 4), (4, 4), True),
      ('batched scale_tri, batched value', (3, 7, 4), (4,), (7, 4, 4), True),
      ('batched scale_tri, batched value, upper',
       (3, 7, 4), (4,), (7, 4, 4), False),
  )
  def test_log_prob(self, value_shape, loc_shape, scale_tri_shape, is_lower):
    rng = np.random.default_rng(2022)
    loc = rng.normal(size=loc_shape)
    scale_tri = rng.normal(size=scale_tri_shape)
    value = rng.normal(size=value_shape)
    dist_kwargs = {'loc': loc, 'scale_tri': scale_tri, 'is_lower': is_lower}
    if is_lower:
      tfp_dist_kwargs = {'loc': loc, 'scale_tril': scale_tri}
    else:
      scale_tril = _get_scale_tril_from_scale_triu(scale_tri)
      tfp_dist_kwargs = {'loc': loc, 'scale_tril': scale_tril}
    super()._test_attribute(
        attribute_string='log_prob',
        dist_kwargs=dist_kwargs,
        tfp_dist_kwargs=tfp_dist_kwargs,
        call_args=(value,),
        assertion_fn=self.assertion_fn(rtol=1e-3))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('unbatched', (4,), (4, 4)),
      ('batched loc', (7, 4), (4, 4)),
      ('batched scale_tri', (4,), (7, 4, 4)),
  )
  def test_method(self, loc_shape, scale_tri_shape):
    rng = np.random.default_rng(2022)
    loc = rng.normal(size=loc_shape)
    scale_tri = rng.normal(size=scale_tri_shape)
    for method in ['entropy', 'mean', 'stddev', 'variance',
                   'covariance', 'mode']:
      for is_lower in [True, False]:
        if method in ['stddev', 'covariance', 'variance']:
          rtol = 2e-2 if is_lower else 5e-2
        else:
          rtol = 1e-3

        dist_kwargs = {'loc': loc, 'scale_tri': scale_tri, 'is_lower': is_lower}
        if is_lower:
          tfp_dist_kwargs = {'loc': loc, 'scale_tril': scale_tri}
        else:
          scale_tril = _get_scale_tril_from_scale_triu(scale_tri)
          tfp_dist_kwargs = {'loc': loc, 'scale_tril': scale_tril}
        with self.subTest(method=method, is_lower=is_lower):
          super()._test_attribute(
              method,
              dist_kwargs=dist_kwargs,
              tfp_dist_kwargs=tfp_dist_kwargs,
              assertion_fn=self.assertion_fn(rtol=rtol))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('kl distrax_to_distrax', 'kl_divergence', 'distrax_to_distrax'),
      ('kl distrax_to_tfp', 'kl_divergence', 'distrax_to_tfp'),
      ('kl tfp_to_distrax', 'kl_divergence', 'tfp_to_distrax'),
      ('cross-ent distrax_to_distrax', 'cross_entropy', 'distrax_to_distrax'),
      ('cross-ent distrax_to_tfp', 'cross_entropy', 'distrax_to_tfp'),
      ('cross-ent tfp_to_distrax', 'cross_entropy', 'tfp_to_distrax'))
  def test_with_two_distributions(self, function_string, mode_string):
    rng = np.random.default_rng(2022)
    loc1 = rng.normal(size=(5, 1, 4))
    scale_tri1 = rng.normal(size=(3, 4, 4))
    loc2 = rng.normal(size=(3, 4))
    scale_tri2 = rng.normal(size=(4, 4))
    for is_lower in [True, False]:
      dist1_kwargs = {
          'loc': loc1, 'scale_tri': scale_tri1, 'is_lower': is_lower}
      dist2_kwargs = {
          'loc': loc2, 'scale_tri': scale_tri2, 'is_lower': is_lower}
      if is_lower:
        tfp_dist1_kwargs = {'loc': loc1, 'scale_tril': scale_tri1}
        tfp_dist2_kwargs = {'loc': loc2, 'scale_tril': scale_tri2}
      else:
        tfp_dist1_kwargs = {
            'loc': loc1,
            'scale_tril': _get_scale_tril_from_scale_triu(scale_tri1)
        }
        tfp_dist2_kwargs = {
            'loc': loc2,
            'scale_tril': _get_scale_tril_from_scale_triu(scale_tri2)
        }
      with self.subTest(is_lower=is_lower):
        super()._test_with_two_distributions(
            attribute_string=function_string,
            mode_string=mode_string,
            dist1_kwargs=dist1_kwargs,
            dist2_kwargs=dist2_kwargs,
            tfp_dist1_kwargs=tfp_dist1_kwargs,
            tfp_dist2_kwargs=tfp_dist2_kwargs,
            assertion_fn=self.assertion_fn(rtol=1e-3))

  def test_jittable(self):
    super()._test_jittable(
        dist_kwargs={'loc': np.zeros((4,))},
        assertion_fn=self.assertion_fn(rtol=1e-3))

  @parameterized.named_parameters(
      ('single element', 2),
      ('range', slice(-1)),
      ('range_2', (slice(None), slice(-1))),
  )
  def test_slice(self, slice_):
    rng = np.random.default_rng(2022)
    loc = rng.normal(size=(6, 5, 4))
    scale_tri = rng.normal(size=(4, 4))
    for is_lower in [True, False]:
      with self.subTest(is_lower=is_lower):
        dist_kwargs = {'loc': loc, 'scale_tri': scale_tri, 'is_lower': is_lower}
        dist = self.distrax_cls(**dist_kwargs)
        self.assertEqual(dist[slice_].batch_shape, loc[slice_].shape[:-1])
        self.assertEqual(dist[slice_].event_shape, dist.event_shape)
        self.assertEqual(dist[slice_].is_lower, dist.is_lower)
        self.assertion_fn(rtol=1e-3)(dist[slice_].mean(), loc[slice_])

  def test_slice_ellipsis(self):
    rng = np.random.default_rng(2022)
    loc = rng.normal(size=(6, 5, 4))
    scale_tri = rng.normal(size=(4, 4))
    for is_lower in [True, False]:
      with self.subTest(is_lower=is_lower):
        dist_kwargs = {'loc': loc, 'scale_tri': scale_tri, 'is_lower': is_lower}
        dist = self.distrax_cls(**dist_kwargs)
        self.assertEqual(dist[..., -1].batch_shape, (6,))
        self.assertEqual(dist[..., -1].event_shape, dist.event_shape)
        self.assertEqual(dist[..., -1].is_lower, dist.is_lower)
        self.assertion_fn(rtol=1e-3)(dist[..., -1].mean(), loc[:, -1, :])


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""MultivariateNormalDiagPlusLowRank distribution."""

from typing import Optional

import chex
from distrax._src.bijectors.diag_linear import DiagLinear
from distrax._src.bijectors.diag_plus_low_rank_linear import DiagPlusLowRankLinear
from distrax._src.distributions import distribution
from distrax._src.distributions.mvn_from_bijector import MultivariateNormalFromBijector
from distrax._src.utils import conversion
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions

Array = chex.Array


def _check_parameters(
    loc: Optional[Array],
    scale_diag: Optional[Array],
    scale_u_matrix: Optional[Array],
    scale_v_matrix: Optional[Array]) -> None:
  """Checks that the inputs are correct."""

  if all(x is None for x in [loc, scale_diag, scale_u_matrix]):
    raise ValueError(
        'At least one of `loc`, `scale_diag`,  and `scale_u_matrix` must '
        'be specified.')
  if scale_v_matrix is not None and scale_u_matrix is None:
    raise ValueError('`scale_v_matrix` can be specified only when '
                     '`scale_u_matrix` is also specified.')

  for name, x, n_dims in [
      ('loc', loc, 1), ('scale_diag', scale_diag, 1),
      ('scale_u_matrix', scale_u_matrix, 2),
      ('scale_v_matrix', scale_v_matrix, 2)]:
    if x is not None and x.ndim < n_dims:
      raise ValueError(f'`{name}` must have at least {n_dims} dimensions.')

  if scale_u_matrix is not None and scale_v_matrix is not None:
    if scale_u_matrix.shape[-1] != scale_v_matrix.shape[-1]:
      raise ValueError(
          f'The last dimension of `scale_u_matrix` must coincide with '
          f'the last dimension of `scale_v_matrix`, but got '
          f'`scale_u_matrix.shape[-1] = {scale_u_matrix.shape[-1]}`'
          f' and `scale_v_matrix.shape[-1] = {scale_v_matrix.shape[-1]}`.')

  if scale_u_matrix is not None and scale_u_matrix.shape[-1] < 1:
    raise ValueError(
        'The last dimension of `scale_u_matrix` cannot be zero.')

  loc_dim = None if loc is None else loc.shape[-1]
  scale_diag_dim = None if scale_diag is None else scale_diag.shape[-1]
  scale_u_matrix_dim = (
      None if scale_u_matrix is None else scale_u_matrix.shape[-2])
  scale_v_matrix_dim = (
      None if scale_v_matrix is None else scale_v_matrix.shape[-2])
  num_dims = loc_dim if loc_dim is not None else scale_diag_dim
  num_dims = num_dims if num_dims is not None else scale_u_matrix_dim
  array_dims = [
      x for x in [
          loc_dim, scale_diag_dim, scale_u_matrix_dim, scale_v_matrix_dim]
      if x is not None]
  if not all(x == num_dims for x in array_dims):
    raise ValueError(
        f'If specified, the following shapes must all coincide, but got '
        f'`loc.shape[-1] = {loc_dim}`, '
        f'`scale_diag.shape[-1] = {scale_diag_dim}`, '
        f'`scale_u_matrix.shape[-2] = {scale_u_matrix_dim}`, and '
        f'`scale_v_matrix.shape[-2] = {scale_v_matrix_dim}`.')


class MultivariateNormalDiagPlusLowRank(MultivariateNormalFromBijector):
  """Multivariate normal distribution on `R^k`.

  The `MultivariateNormalDiagPlusLowRank` distribution is parameterized by a
  location (mean) vector `b` and a scale matrix `S` that has the following
  structure: `S = diag(D) + U @ V.T`, where `D` is a `k`-length vector, and both
  `U` and `V` are `k x r` matrices (with `r < k` typically). The covariance
  matrix of the multivariate normal distribution is `C = S @ S.T`.

  This class makes no attempt to verify that the scale matrix `S` is invertible,
  which happens if and only if both `diag(D)` and `I + V^T diag(D)^{-1} U` are
  invertible. It is the responsibility of the user to make sure that this is the
  case.
  """

  equiv_tfp_cls = tfd.MultivariateNormalDiagPlusLowRank

  def __init__(self,
               loc: Optional[Array] = None,
               scale_diag: Optional[Array] = None,
               scale_u_matrix: Optional[Array] = None,
               scale_v_matrix: Optional[Array] = None):
    """Initializes a MultivariateNormalDiagPlusLowRank distribution.

    Args:
      loc: Mean vector of the distribution of shape `k` (can also be a batch of
        such vectors). If not specified, it defaults to zeros.
      scale_diag: The diagonal matrix added to the scale `S`, specified by a
        `k`-length vector containing its diagonal entries (or a batch of
        vectors). If not specified, the diagonal matrix defaults to the
        identity.
      scale_u_matrix: The low-rank matrix `U` that specifies the scale, as
        described in the class docstring. It is a `k x r` matrix (or a batch of
        such matrices). If not specified, it defaults to zeros. At least one
        of `loc`, `scale_diag`, and `scale_u_matrix` must be specified.
      scale_v_matrix: The low-rank matrix `V` that specifies the scale, as
        described in the class docstring. It is a `k x r` matrix (or a batch of
        such matrices). If not specified, it defaults to `scale_u_matrix`. It
        can only be specified if `scale_u_matrix` is also specified.
    """
    loc = None if loc is None else conversion.as_float_array(loc)
    scale_diag = None if scale_diag is None else conversion.as_float_array(
        scale_diag)
    scale_u_matrix = (
        None if scale_u_matrix is None else conversion.as_float_array(
            scale_u_matrix))
    scale_v_matrix = (
        None if scale_v_matrix is None else conversion.as_float_array(
            scale_v_matrix))

    _check_parameters(loc, scale_diag, scale_u_matrix, scale_v_matrix)

    num_dims = None
    if loc is not None:
      num_dims = loc.shape[-1]
    elif scale_diag is not None:
      num_dims = scale_diag.shape[-1]
    elif scale_u_matrix is not None:
      num_dims = scale_u_matrix.shape[-2]

    dtype = jnp.result_type(
        *[x for x in [loc, scale_diag, scale_u_matrix, scale_v_matrix]
          if x is not None])

    if loc is None:
      assert num_dims is not None
      loc = jnp.zeros((num_dims,), dtype=dtype)

    self._scale_diag = scale_diag
    if scale_diag is None:
      self._scale_diag = jnp.ones((num_dims,), dtype=dtype)

    self._scale_u_matrix = scale_u_matrix
    if scale_u_matrix is None:
      self._scale_u_matrix = jnp.zeros((num_dims, 1), dtype=dtype)

    self._scale_v_matrix = scale_v_matrix
    if scale_v_matrix is None:
      self._scale_v_matrix = self._scale_u_matrix

    if scale_u_matrix is None:
      # The scale matrix is diagonal.
      scale = DiagLinear(self._scale_diag)
    else:
      scale = DiagPlusLowRankLinear(
          u_matrix=self._scale_u_matrix,
          v_matrix=self._scale_v_matrix,
          diag=self._scale_diag)
    super().__init__(loc=loc, scale=scale)

  @property
  def scale_diag(self) -> Array:
    """Diagonal matrix that is added to the scale."""
    return jnp.broadcast_to(
        self._scale_diag, self.batch_shape + self.event_shape)

  @property
  def scale_u_matrix(self) -> Array:
    """Matrix `U` that defines the low-rank part of the scale matrix."""
    return jnp.broadcast_to(
        self._scale_u_matrix,
        self.batch_shape + self._scale_u_matrix.shape[-2:])

  @property
  def scale_v_matrix(self) -> Array:
    """Matrix `V` that defines the low-rank part of the scale matrix."""
    return jnp.broadcast_to(
        self._scale_v_matrix,
        self.batch_shape + self._scale_v_matrix.shape[-2:])

  def __getitem__(self, index) -> 'MultivariateNormalDiagPlusLowRank':
    """See `Distribution.__getitem__`."""
    index = distribution.to_batch_shape_index(self.batch_shape, index)
    return MultivariateNormalDiagPlusLowRank(
        loc=self.loc[index],
        scale_diag=self.scale_diag[index],
        scale_u_matrix=self.scale_u_matrix[index],
        scale_v_matrix=self.scale_v_matrix[index])



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `mvn_diag.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions import mvn_diag
from distrax._src.distributions import normal
from distrax._src.utils import equivalence
import jax
import jax.experimental
import jax.numpy as jnp
import numpy as np


class MultivariateNormalDiagTest(equivalence.EquivalenceTest):

  def setUp(self):
    super().setUp()
    self._init_distr_cls(mvn_diag.MultivariateNormalDiag)

  @parameterized.named_parameters(
      ('1d std normal', {'scale_diag': np.ones((1,))}),
      ('2d std normal', {'loc': [0., 0.], 'scale_diag': [1., 1.]}),
      ('2d std normal, None loc', {'scale_diag': [1., 1.]}),
      ('2d std normal, None scale_diag', {'loc': [0., 0.]}),
      ('rank-2 parameters',
       {'loc': np.zeros((3, 2)), 'scale_diag': np.ones((3, 2))}),
      ('broadcasted scale_diag',
       {'loc': np.zeros((3, 2)), 'scale_diag': np.ones((2,))}),
      ('broadcasted loc',
       {'loc': np.zeros((2)), 'scale_diag': np.ones((3, 2,))}),
  )
  def test_event_shape(self, distr_params):
    distr_params = {
        k: np.asarray(v, dtype=np.float32) for k, v in distr_params.items()}
    super()._test_event_shape((), distr_params)

  @parameterized.named_parameters(
      ('1d std normal', {'scale_diag': np.ones((1,))}),
      ('2d std normal', {'loc': [0., 0.], 'scale_diag': [1., 1.]}),
      ('2d std normal, None loc', {'scale_diag': [1., 1.]}),
      ('2d std normal, None scale_diag', {'loc': [0., 0.]}),
      ('rank-2 parameters',
       {'loc': np.zeros((3, 2)), 'scale_diag': np.ones((3, 2))}),
      ('broadcasted scale_diag',
       {'loc': np.zeros((3, 2)), 'scale_diag': np.ones((2,))}),
      ('broadcasted loc',
       {'loc': np.zeros((2)), 'scale_diag': np.ones((3, 2,))}),
  )
  def test_batch_shape(self, distr_params):
    distr_params = {
        k: np.asarray(v, dtype=np.float32) for k, v in distr_params.items()}
    super()._test_batch_shape((), distr_params)

  def test_invalid_parameters(self):
    self._test_raises_error(dist_kwargs={'loc': None, 'scale_diag': None})
    self._test_raises_error(
        dist_kwargs={'loc': None, 'scale_diag': np.array(1.)})
    self._test_raises_error(
        dist_kwargs={'loc': np.array(1.), 'scale_diag': None})
    self._test_raises_error(
        dist_kwargs={'loc': np.zeros((3, 5)), 'scale_diag': np.ones((3, 4))})

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d std normal, no shape',
       {'scale_diag': np.ones((1,))},
       ()),
      ('2d std normal, no shape',
       {'loc': [0., 0.],
        'scale_diag': [1., 1.]},
       ()),
      ('2d std normal, None loc, no shape',
       {'scale_diag': [1., 1.]},
       ()),
      ('2d std normal, None scale_diag, no shape',
       {'loc': [0., 0.]},
       ()),
      ('2d std normal, int shape',
       {'loc': [0., 0.],
        'scale_diag': [1., 1.]},
       3),
      ('2d std normal, None loc, int shape',
       {'scale_diag': [1., 1.]},
       3),
      ('2d std normal, None scale_diag, int shape',
       {'loc': [0., 0.]},
       3),
      ('2d std normal, 1-tuple shape',
       {'loc': [0., 0.],
        'scale_diag': [1., 1.]},
       (3,)),
      ('2d std normal, None loc, 1-tuple shape',
       {'scale_diag': [1., 1.]},
       (3,)),
      ('2d std normal, None scale_diag, 1-tuple shape',
       {'loc': [0., 0.]},
       (3,)),
      ('2d std normal, 2-tuple shape',
       {'loc': [0., 0.],
        'scale_diag': [1., 1.]},
       (3, 4)),
      ('2d std normal, None loc, 2-tuple shape',
       {'scale_diag': [1., 1.]},
       (3, 4)),
      ('2d std normal, None scale_diag, 2-tuple shape',
       {'loc': [0., 0.]},
       (3, 4)),
      ('rank-2 parameters, 2-tuple shape',
       {'loc': np.zeros((3, 2)),
        'scale_diag': np.ones((3, 2))},
       (5, 4)),
      ('broadcasted scale_diag',
       {'loc': np.zeros((3, 2)),
        'scale_diag': np.ones((2,))},
       5),
      ('broadcasted loc',
       {'loc': np.zeros((2)),
        'scale_diag': np.ones((3, 2,))},
       5),
  )
  def test_sample_shape(self, distr_params, sample_shape):
    distr_params = {
        k: np.asarray(v, dtype=np.float32) for k, v in distr_params.items()}
    super()._test_sample_shape(
        dist_args=(),
        dist_kwargs=distr_params,
        sample_shape=sample_shape)

  @chex.all_variants
  @jax.numpy_rank_promotion('raise')
  @parameterized.named_parameters(
      ('1d std normal, no shape',
       {'scale_diag': np.ones((1,))},
       ()),
      ('2d std normal, no shape',
       {'loc': [0., 0.],
        'scale_diag': [1., 1.]},
       ()),
      ('2d std normal, None loc, no shape',
       {'scale_diag': [1., 1.]},
       ()),
      ('2d std normal, None scale_diag, no shape',
       {'loc': [0., 0.]},
       ()),
      ('2d std normal, int shape',
       {'loc': [0., 0.],
        'scale_diag': [1., 1.]},
       3),
      ('2d std normal, None loc, int shape',
       {'scale_diag': [1., 1.]},
       3),
      ('2d std normal, None scale_diag, int shape',
       {'loc': [0., 0.]},
       3),
      ('2d std normal, 1-tuple shape',
       {'loc': [0., 0.],
        'scale_diag': [1., 1.]},
       (3,)),
      ('2d std normal, None loc, 1-tuple shape',
       {'scale_diag': [1., 1.]},
       (3,)),
      ('2d std normal, None scale_diag, 1-tuple shape',
       {'loc': [0., 0.]},
       (3,)),
      ('2d std normal, 2-tuple shape',
       {'loc': [0., 0.],
        'scale_diag': [1., 1.]},
       (3, 4)),
      ('2d std normal, None loc, 2-tuple shape',
       {'scale_diag': [1., 1.]},
       (3, 4)),
      ('2d std normal, None scale_diag, 2-tuple shape',
       {'loc': [0., 0.]},
       (3, 4)),
      ('rank-2 parameters, 2-tuple shape',
       {'loc': np.zeros((3, 2)),
        'scale_diag': np.ones((3, 2))},
       (5, 4)),
      ('broadcasted scale_diag',
       {'loc': np.zeros((3, 2)),
        'scale_diag': np.ones((2,))},
       5),
      ('broadcasted loc',
       {'loc': np.zeros((2)),
        'scale_diag': np.ones((3, 2,))},
       5),
  )
  def test_sample_and_log_prob(self, distr_params, sample_shape):
    distr_params = {
        k: np.asarray(v, dtype=np.float32) for k, v in distr_params.items()}
    super()._test_sample_and_log_prob(
        dist_args=(),
        dist_kwargs=distr_params,
        sample_shape=sample_shape,
        assertion_fn=self.assertion_fn(rtol=1e-3))

  @chex.all_variants
  @parameterized.named_parameters(
      ('float32', jnp.float32),
      ('float64', jnp.float64))
  def test_sample_dtype(self, dtype):
    with jax.experimental.enable_x64(dtype.dtype.itemsize == 8):
      dist_params = {
          'loc': np.array([0., 0.], dtype),
          'scale_diag': np.array([1., 1.], dtype)}
      dist = self.distrax_cls(**dist_params)
      samples = self.variant(dist.sample)(seed=self.key)
      self.assertEqual(samples.dtype, dist.dtype)
      chex.assert_type(samples, dtype)

  @chex.all_variants
  @parameterized.named_parameters(
      ('log_prob; 2d dist, 2d value', 'log_prob',
       {'scale_diag': [1., 1.]},
       [0., -0.5]),
      ('log_prob; 3d dist, broadcasted params, 3d value', 'log_prob',
       {'loc': np.zeros((4, 3)),
        'scale_diag': 0.3 * np.ones((3,))},
       [-0.1, 0., -0.5]),
      ('log_prob; 3d dist, broadcasted scale_diag, rank-2 value', 'log_prob',
       {'loc': np.zeros((4, 3)),
        'scale_diag': 0.1 * np.ones((3,))},
       -0.1 * np.ones((4, 3))),
      ('log_prob; 3d dist, broadcasted scale_diag, rank-3 value', 'log_prob',
       {'loc': np.zeros((4, 3)),
        'scale_diag': 0.1 * np.ones((3,))},
       -0.1 * np.ones((5, 4, 3))),
      ('log_prob; 2d dist, 2d value, edge case', 'log_prob',
       {'scale_diag': [1., 1.]},
       [200., -200.]),
      ('prob; 2d dist, 2d value', 'prob',
       {'scale_diag': [1., 1.]},
       [0., -0.5]),
      ('prob; 3d dist, broadcasted params, 3d value', 'prob',
       {'loc': np.zeros((4, 3)),
        'scale_diag': 0.3 * np.ones((3,))},
       [-0.1, 0., -0.5]),
      ('prob; 3d dist, broadcasted scale_diag, rank-2 value', 'prob',
       {'loc': np.zeros((4, 3)),
        'scale_diag': 0.1 * np.ones((3,))},
       -0.1 * np.ones((4, 3))),
      ('prob; 3d dist, broadcasted scale_diag, rank-3 value', 'prob',
       {'loc': np.zeros((4, 3)),
        'scale_diag': 0.1 * np.ones((3,))},
       -0.1 * np.ones((5, 4, 3))),
      ('prob; 2d dist, 2d value, edge case', 'prob',
       {'scale_diag': [1., 1.]},
       [200., -200.]),
  )
  def test_pdf(self, function_string, distr_params, value):
    distr_params = {
        k: np.asarray(v, dtype=np.float32) for k, v in distr_params.items()}
    value = np.asarray(value)
    super()._test_attribute(
        attribute_string=function_string,
        dist_kwargs=distr_params,
        call_args=(value,),
        assertion_fn=self.assertion_fn(rtol=1e-3))

  @chex.all_variants
  @parameterized.named_parameters(
      ('log_cdf; 2d dist, 2d value', 'log_cdf',
       {'scale_diag': [1., 1.]},
       [0., -0.5]),
      ('log_cdf; 3d dist, broadcasted params, 3d value', 'log_cdf',
       {'loc': np.zeros((4, 3)),
        'scale_diag': 0.3 * np.ones((3,))},
       [-0.1, 0., -0.5]),
      ('log_cdf; 3d dist, broadcasted scale_diag, rank-2 value', 'log_cdf',
       {'loc': np.zeros((4, 3)),
        'scale_diag': 0.1 * np.ones((3,))},
       -0.1 * np.ones((4, 3))),
      ('log_cdf; 3d dist, broadcasted scale_diag, rank-3 value', 'log_cdf',
       {'loc': np.zeros((4, 3)),
        'scale_diag': 0.1 * np.ones((3,))},
       -0.1 * np.ones((5, 4, 3))),
      ('log_cdf; 2d dist, 2d value, edge case', 'log_cdf',
       {'scale_diag': [1., 1.]},
       [200., -200.]),
      ('cdf; 2d dist, 2d value', 'cdf',
       {'scale_diag': [1., 1.]},
       [0., -0.5]),
      ('cdf; 3d dist, broadcasted params, 3d value', 'cdf',
       {'loc': np.zeros((4, 3)),
        'scale_diag': 0.3 * np.ones((3,))},
       [-0.1, 0., -0.5]),
      ('cdf; 3d dist, broadcasted scale_diag, rank-2 value', 'cdf',
       {'loc': np.zeros((4, 3)),
        'scale_diag': 0.1 * np.ones((3,))},
       -0.1 * np.ones((4, 3))),
      ('cdf; 3d dist, broadcasted scale_diag, rank-3 value', 'cdf',
       {'loc': np.zeros((4, 3)),
        'scale_diag': 0.1 * np.ones((3,))},
       -0.1 * np.ones((5, 4, 3))),
      ('cdf; 2d dist, 2d value, edge case', 'cdf',
       {'scale_diag': [1., 1.]},
       [200., -200.]),
  )
  def test_cdf(self, function_string, distr_params, value):
    distr_params = {
        k: np.asarray(v, dtype=np.float32) for k, v in distr_params.items()}
    value = np.asarray(value)
    dist = self.distrax_cls(**distr_params)
    result = self.variant(getattr(dist, function_string))(value)
    # The `cdf` is not implemented in TFP, so we test against a `Normal`.
    loc = 0. if 'loc' not in distr_params else distr_params['loc']
    univariate_normal = normal.Normal(loc, distr_params['scale_diag'])
    expected_result = getattr(univariate_normal, function_string)(value)
    if function_string == 'cdf':
      reduce_fn = lambda x: jnp.prod(x, axis=-1)
    elif function_string == 'log_cdf':
      reduce_fn = lambda x: jnp.sum(x, axis=-1)
    else:
      raise ValueError(f'Unsupported function: {function_string}')
    expected_result = reduce_fn(expected_result)
    self.assertion_fn(rtol=1e-3)(result, expected_result)

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('entropy; one distribution', 'entropy',
       {'loc': [0.1, -0.1],
        'scale_diag': [0.8, 0.5]}),
      ('entropy; broadcasted loc', 'entropy',
       {'loc': [0., 0.1, -0.1],
        'scale_diag': [[1.5, 0.8, 0.5], [0.8, 0.1, 0.4]]}),
      ('entropy; broadcasted scale_diag', 'entropy',
       {'loc': [[0., 0.1, -0.1], [0.2, 0., 0.5]],
        'scale_diag': [1.5, 0.8, 0.5]}),
      ('entropy; None loc', 'entropy',
       {'scale_diag': [0.8, 0.5]}),
      ('entropy; None scale_diag', 'entropy',
       {'loc': [0.1, -0.1]}),
      ('mean; one distribution', 'mean',
       {'loc': [0.1, -0.1],
        'scale_diag': [0.8, 0.5]}),
      ('mean; broadcasted loc', 'mean',
       {'loc': [0., 0.1, -0.1],
        'scale_diag': [[1.5, 0.8, 0.5], [0.8, 0.1, 0.4]]}),
      ('mean; broadcasted scale_diag', 'mean',
       {'loc': [[0., 0.1, -0.1], [0.2, 0., 0.5]],
        'scale_diag': [1.5, 0.8, 0.5]}),
      ('mean; None loc', 'mean',
       {'scale_diag': [0.8, 0.5]}),
      ('mean; None scale_diag', 'mean',
       {'loc': [0.1, -0.1]}),
      ('stddev; one distribution', 'stddev',
       {'loc': [0.1, -0.1],
        'scale_diag': [0.8, 0.5]}),
      ('stddev; broadcasted loc', 'stddev',
       {'loc': [0., 0.1, -0.1],
        'scale_diag': [[1.5, 0.8, 0.5], [0.8, 0.1, 0.4]]}),
      ('stddev; broadcasted scale_diag', 'stddev',
       {'loc': [[0., 0.1, -0.1], [0.2, 0., 0.5]],
        'scale_diag': [1.5, 0.8, 0.5]}),
      ('stddev; None loc', 'stddev',
       {'scale_diag': [0.8, 0.5]}),
      ('stddev; None scale_diag', 'stddev',
       {'loc': [0.1, -0.1]}),
      ('variance; one distribution', 'variance',
       {'loc': [0.1, -0.1],
        'scale_diag': [0.8, 0.5]}),
      ('variance; broadcasted loc', 'variance',
       {'loc': [0., 0.1, -0.1],
        'scale_diag': [[1.5, 0.8, 0.5], [0.8, 0.1, 0.4]]}),
      ('variance; broadcasted scale_diag', 'variance',
       {'loc': [[0., 0.1, -0.1], [0.2, 0., 0.5]],
        'scale_diag': [1.5, 0.8, 0.5]}),
      ('variance; None loc', 'variance',
       {'scale_diag': [0.8, 0.5]}),
      ('variance; None scale_diag', 'variance',
       {'loc': [0.1, -0.1]}),
      ('covariance; one distribution', 'covariance',
       {'loc': [0.1, -0.1],
        'scale_diag': [0.8, 0.5]}),
      ('covariance; broadcasted loc', 'covariance',
       {'loc': [0., 0.1, -0.1],
        'scale_diag': [[1.5, 0.8, 0.5], [0.8, 0.1, 0.4]]}),
      ('covariance; None loc', 'covariance',
       {'scale_diag': [0.8, 0.5]}),
      ('covariance; None scale_diag', 'covariance',
       {'loc': [0.1, -0.1]}),
      ('mode; broadcasted scale_diag', 'mode',
       {'loc': [[0., 0.1, -0.1], [0.2, 0., 0.5]],
        'scale_diag': [1.5, 0.8, 0.5]}),
  )
  def test_method(self, function_string, distr_params):
    distr_params = {
        k: np.asarray(v, dtype=np.float32) for k, v in distr_params.items()}
    super()._test_attribute(
        function_string,
        dist_kwargs=distr_params,
        assertion_fn=self.assertion_fn(rtol=1e-3))

  @chex.all_variants(with_pmap=False)
  def test_median(self):
    dist_params = {'loc': np.array([0.3, -0.1, 0.0]),
                   'scale_diag': np.array([0.1, 1.4, 0.5])}
    dist = self.distrax_cls(**dist_params)
    self.assertion_fn(rtol=1e-3)(self.variant(dist.median)(), dist.mean())

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('kl distrax_to_distrax', 'kl_divergence', 'distrax_to_distrax'),
      ('kl distrax_to_tfp', 'kl_divergence', 'distrax_to_tfp'),
      ('kl tfp_to_distrax', 'kl_divergence', 'tfp_to_distrax'),
      ('cross-ent distrax_to_distrax', 'cross_entropy', 'distrax_to_distrax'),
      ('cross-ent distrax_to_tfp', 'cross_entropy', 'distrax_to_tfp'),
      ('cross-ent tfp_to_distrax', 'cross_entropy', 'tfp_to_distrax'))
  def test_with_two_distributions(self, function_string, mode_string):
    rng = np.random.default_rng(42)
    super()._test_with_two_distributions(
        attribute_string=function_string,
        mode_string=mode_string,
        dist1_kwargs={
            'loc': rng.normal(size=(4, 1, 5)).astype(np.float32),
            'scale_diag': 0.1 + rng.uniform(size=(3, 5)).astype(np.float32),
        },
        dist2_kwargs={
            'loc': np.asarray([-2.4, -1., 0., 1.2, 6.5]).astype(np.float32),
            'scale_diag': None,
        },
        assertion_fn=self.assertion_fn(rtol=1e-3))

  def test_jittable(self):
    super()._test_jittable(
        (np.zeros((2, 3,)), np.ones((2, 3,))),
        assertion_fn=self.assertion_fn(rtol=1e-3))

  @parameterized.named_parameters(
      ('single element', 2),
      ('range', slice(-1)),
      ('range_2', (slice(None), slice(-1))),
  )
  def test_slice(self, slice_):
    rng = np.random.default_rng(42)
    loc = jnp.array(rng.normal(size=(3, 4, 5)))
    scale_diag = jnp.array(rng.uniform(size=(3, 4, 5)))
    dist = self.distrax_cls(loc=loc, scale_diag=scale_diag)
    self.assertion_fn(rtol=1e-3)(dist[slice_].mean(), loc[slice_])

  def test_slice_different_parameterization(self):
    rng = np.random.default_rng(42)
    loc = jnp.array(rng.normal(size=(4,)))
    scale_diag = jnp.array(rng.uniform(size=(3, 4)))
    dist = self.distrax_cls(loc=loc, scale_diag=scale_diag)
    self.assertion_fn(rtol=1e-3)(dist[0].mean(), loc)  # Not slicing loc.
    self.assertion_fn(rtol=1e-3)(dist[0].stddev(), scale_diag[0])

  def test_slice_ellipsis(self):
    rng = np.random.default_rng(42)
    loc = jnp.array(rng.normal(size=(3, 4, 5)))
    scale_diag = jnp.array(rng.uniform(size=(3, 4, 5)))
    dist = self.distrax_cls(loc=loc, scale_diag=scale_diag)
    self.assertion_fn(rtol=1e-3)(dist[..., -1].mean(), loc[:, -1])
    self.assertion_fn(rtol=1e-3)(dist[..., -1].stddev(), scale_diag[:, -1])


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `split_coupling.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.bijectors import bijector as base_bijector
from distrax._src.bijectors import block
from distrax._src.bijectors import split_coupling
import jax
import jax.numpy as jnp
import numpy as np


def _create_split_coupling_bijector(split_index,
                                    split_axis=-1,
                                    swap=False,
                                    event_ndims=2):
  return split_coupling.SplitCoupling(
      split_index=split_index,
      split_axis=split_axis,
      event_ndims=event_ndims,
      swap=swap,
      conditioner=lambda x: x**2,
      bijector=lambda _: lambda x: 2. * x + 3.)


class DummyBijector(base_bijector.Bijector):

  def forward_and_log_det(self, x):
    super()._check_forward_input_shape(x)
    return x, jnp.zeros((x.shape[:-self.event_ndims_in]), dtype=jnp.float_)


class SplitCouplingTest(parameterized.TestCase):

  def test_properties(self):
    bijector = _create_split_coupling_bijector(
        split_index=0, swap=False, split_axis=-1, event_ndims=2)
    ones = jnp.ones((4, 5))
    self.assertEqual(bijector.split_index, 0)
    self.assertEqual(bijector.split_axis, -1)
    self.assertFalse(bijector.swap)
    np.testing.assert_allclose(
        bijector.conditioner(2 * ones), 4 * ones, atol=1e-4)
    assert callable(bijector.bijector(ones))

  @parameterized.named_parameters(
      ('negative split_index', {'split_index': -1, 'event_ndims': 0}),
      ('positive split_axis',
       {'split_index': 0, 'event_ndims': 0, 'split_axis': 3}),
      ('negative event_ndims', {'split_index': 0, 'event_ndims': -1}),
      ('invalid split_axis',
       {'split_index': 0, 'event_ndims': 1, 'split_axis': -2}),
  )
  def test_invalid_properties(self, bij_params):
    bij_params.update(
        {'conditioner': lambda x: x, 'bijector': lambda _: lambda x: x})
    with self.assertRaises(ValueError):
      split_coupling.SplitCoupling(**bij_params)

  def test_raises_on_bijector_with_different_event_ndims(self):
    inner_bij = lambda _: DummyBijector(1, 0, False, False)
    bij_params = {'split_index': 0, 'event_ndims': 1,
                  'conditioner': lambda x: x, 'bijector': inner_bij}
    bij = split_coupling.SplitCoupling(**bij_params)
    with self.assertRaises(ValueError):
      bij.forward_and_log_det(jnp.zeros((4, 3)))

  def test_raises_on_bijector_with_extra_event_ndims(self):
    inner_bij = lambda _: DummyBijector(2, 2, False, False)
    bij_params = {'split_index': 0, 'event_ndims': 1,
                  'conditioner': lambda x: x, 'bijector': inner_bij}
    bij = split_coupling.SplitCoupling(**bij_params)
    with self.assertRaises(ValueError):
      bij.forward_and_log_det(jnp.zeros((4, 3)))

  @chex.all_variants
  @parameterized.parameters(
      {'split_index': 0, 'split_axis': -1, 'swap': False},
      {'split_index': 3, 'split_axis': -1, 'swap': False},
      {'split_index': 5, 'split_axis': -1, 'swap': False},
      {'split_index': 0, 'split_axis': -2, 'swap': False},
      {'split_index': 2, 'split_axis': -2, 'swap': False},
      {'split_index': 4, 'split_axis': -2, 'swap': False},
      {'split_index': 0, 'split_axis': -1, 'swap': True},
      {'split_index': 3, 'split_axis': -1, 'swap': True},
      {'split_index': 5, 'split_axis': -1, 'swap': True},
      {'split_index': 0, 'split_axis': -2, 'swap': True},
      {'split_index': 2, 'split_axis': -2, 'swap': True},
      {'split_index': 4, 'split_axis': -2, 'swap': True},
  )
  def test_shapes_are_correct(self, split_index, split_axis, swap):
    key = jax.random.PRNGKey(42)
    x = jax.random.normal(key, (2, 3, 4, 5))
    bijector = _create_split_coupling_bijector(
        split_index, split_axis, swap, event_ndims=2)
    # Forward methods.
    y, logdet = self.variant(bijector.forward_and_log_det)(x)
    self.assertEqual(y.shape, (2, 3, 4, 5))
    self.assertEqual(logdet.shape, (2, 3))
    # Inverse methods.
    x, logdet = self.variant(bijector.inverse_and_log_det)(y)
    self.assertEqual(x.shape, (2, 3, 4, 5))
    self.assertEqual(logdet.shape, (2, 3))

  @chex.all_variants
  def test_swapping_works(self):
    key = jax.random.PRNGKey(42)
    x = jax.random.normal(key, (2, 3, 4, 5))
    # Don't swap.
    bijector = _create_split_coupling_bijector(
        split_index=3, split_axis=-1, swap=False)
    y = self.variant(bijector.forward)(x)
    np.testing.assert_array_equal(y[..., :3], x[..., :3])
    # Swap.
    bijector = _create_split_coupling_bijector(
        split_index=3, split_axis=-1, swap=True)
    y = self.variant(bijector.forward)(x)
    np.testing.assert_array_equal(y[..., 3:], x[..., 3:])
    # Don't swap.
    bijector = _create_split_coupling_bijector(
        split_index=3, split_axis=-2, swap=False)
    y = self.variant(bijector.forward)(x)
    np.testing.assert_array_equal(y[..., :3, :], x[..., :3, :])
    # Swap.
    bijector = _create_split_coupling_bijector(
        split_index=3, split_axis=-2, swap=True)
    y = self.variant(bijector.forward)(x)
    np.testing.assert_array_equal(y[..., 3:, :], x[..., 3:, :])

  @chex.all_variants
  @parameterized.parameters(
      {'split_index': 0, 'split_axis': -1, 'swap': False},
      {'split_index': 3, 'split_axis': -1, 'swap': False},
      {'split_index': 5, 'split_axis': -1, 'swap': False},
      {'split_index': 0, 'split_axis': -2, 'swap': False},
      {'split_index': 2, 'split_axis': -2, 'swap': False},
      {'split_index': 4, 'split_axis': -2, 'swap': False},
      {'split_index': 0, 'split_axis': -1, 'swap': True},
      {'split_index': 3, 'split_axis': -1, 'swap': True},
      {'split_index': 5, 'split_axis': -1, 'swap': True},
      {'split_index': 0, 'split_axis': -2, 'swap': True},
      {'split_index': 2, 'split_axis': -2, 'swap': True},
      {'split_index': 4, 'split_axis': -2, 'swap': True},
  )
  def test_inverse_methods_are_correct(self, split_index, split_axis, swap):
    key = jax.random.PRNGKey(42)
    x = jax.random.normal(key, (2, 3, 4, 5))
    bijector = _create_split_coupling_bijector(
        split_index, split_axis, swap, event_ndims=2)
    y, logdet_fwd = self.variant(bijector.forward_and_log_det)(x)
    x_rec, logdet_inv = self.variant(bijector.inverse_and_log_det)(y)
    np.testing.assert_allclose(x_rec, x, atol=1e-6)
    np.testing.assert_allclose(logdet_fwd, -logdet_inv, atol=1e-6)

  @chex.all_variants
  @parameterized.parameters(
      {'split_index': 0, 'split_axis': -1, 'swap': False},
      {'split_index': 3, 'split_axis': -1, 'swap': False},
      {'split_index': 5, 'split_axis': -1, 'swap': False},
      {'split_index': 0, 'split_axis': -2, 'swap': False},
      {'split_index': 2, 'split_axis': -2, 'swap': False},
      {'split_index': 4, 'split_axis': -2, 'swap': False},
      {'split_index': 0, 'split_axis': -1, 'swap': True},
      {'split_index': 3, 'split_axis': -1, 'swap': True},
      {'split_index': 5, 'split_axis': -1, 'swap': True},
      {'split_index': 0, 'split_axis': -2, 'swap': True},
      {'split_index': 2, 'split_axis': -2, 'swap': True},
      {'split_index': 4, 'split_axis': -2, 'swap': True},
  )
  def test_composite_methods_are_consistent(self, split_index, split_axis,
                                            swap):
    key = jax.random.PRNGKey(42)
    bijector = _create_split_coupling_bijector(
        split_index, split_axis, swap, event_ndims=2)
    # Forward methods.
    x = jax.random.normal(key, (2, 3, 4, 5))
    y1 = self.variant(bijector.forward)(x)
    logdet1 = self.variant(bijector.forward_log_det_jacobian)(x)
    y2, logdet2 = self.variant(bijector.forward_and_log_det)(x)
    np.testing.assert_allclose(y1, y2, atol=1e-8)
    np.testing.assert_allclose(logdet1, logdet2, atol=1e-8)
    # Inverse methods.
    y = jax.random.normal(key, (2, 3, 4, 5))
    x1 = self.variant(bijector.inverse)(y)
    logdet1 = self.variant(bijector.inverse_log_det_jacobian)(y)
    x2, logdet2 = self.variant(bijector.inverse_and_log_det)(y)
    np.testing.assert_allclose(x1, x2, atol=1e-8)
    np.testing.assert_allclose(logdet1, logdet2, atol=1e-8)

  def test_raises_on_invalid_input_shape(self):
    event_shape = (2, 3)
    bij = split_coupling.SplitCoupling(
        split_index=event_shape[-1] // 2,
        event_ndims=len(event_shape),
        conditioner=lambda x: x,
        bijector=lambda _: lambda x: x)
    for fn in [bij.forward, bij.inverse,
               bij.forward_log_det_jacobian, bij.inverse_log_det_jacobian,
               bij.forward_and_log_det, bij.inverse_and_log_det]:
      with self.assertRaises(ValueError):
        fn(jnp.zeros((3,)))

  def test_raises_on_invalid_inner_bijector(self):
    event_shape = (2, 3)
    bij = split_coupling.SplitCoupling(
        split_index=event_shape[-1] // 2,
        event_ndims=len(event_shape),
        conditioner=lambda x: x,
        bijector=lambda _: block.Block(lambda x: x, len(event_shape) + 1))
    for fn in [bij.forward, bij.inverse,
               bij.forward_log_det_jacobian, bij.inverse_log_det_jacobian,
               bij.forward_and_log_det, bij.inverse_and_log_det]:
      with self.assertRaises(ValueError):
        fn(jnp.zeros(event_shape))

  def test_jittable(self):
    @jax.jit
    def f(x, b):
      return b.forward(x)

    bijector = _create_split_coupling_bijector(0, -1, False, event_ndims=2)
    x = np.zeros((2, 3, 4, 5))
    f(x, bijector)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `chain.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.bijectors import bijector as base_bijector
from distrax._src.bijectors import block
from distrax._src.bijectors import chain
from distrax._src.bijectors import scalar_affine
from distrax._src.distributions import normal
from distrax._src.distributions import transformed
from distrax._src.utils import conversion
import jax
import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp


tfd = tfp.distributions
tfb = tfp.bijectors


RTOL = 1e-2


def _with_additional_parameters(params, all_named_parameters):
  """Convenience function for appending a cartesian product of parameters."""
  for name, param in params:
    for named_params in all_named_parameters:
      yield (f'{named_params[0]}; {name}',) + named_params[1:] + (param,)


def _with_base_dists(*all_named_parameters):
  """Partial of _with_additional_parameters to specify distrax and TFP base."""
  base_dists = (
      ('tfp_base', tfd.Normal),
      ('distrax_base', normal.Normal),
  )
  return _with_additional_parameters(base_dists, all_named_parameters)


class ChainTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.seed = jax.random.PRNGKey(1234)

  def test_properties(self):
    bijector = chain.Chain([tfb.Scale(2), tfb.Shift(3), jnp.tanh])
    for bij in bijector.bijectors:
      assert isinstance(bij, base_bijector.Bijector)

  @parameterized.named_parameters(_with_base_dists(
      ('1d std normal', 0, 1),
      ('2d std normal', np.zeros(2), np.ones(2)),
      ('broadcasted loc', 0, np.ones(3)),
      ('broadcasted scale', np.ones(3), 1),
  ))
  def test_event_shape(self, mu, sigma, base_dist):
    base = base_dist(mu, sigma)
    bijector = chain.Chain([tfb.Scale(2), tfb.Shift(3), tfb.Tanh()])
    dist = transformed.Transformed(base, bijector)

    tfp_bijector = tfb.Chain([tfb.Scale(2), tfb.Shift(3), tfb.Tanh()])
    tfp_dist = tfd.TransformedDistribution(
        conversion.to_tfp(base), tfp_bijector)

    assert dist.event_shape == tfp_dist.event_shape

  @chex.all_variants
  @parameterized.named_parameters(_with_base_dists(
      ('1d std normal, no shape', 0, 1, ()),
      ('1d std normal, int shape', 0, 1, 1),
      ('1d std normal, 1-tuple shape', 0, 1, (1,)),
      ('1d std normal, 2-tuple shape', 0, 1, (2, 2)),
      ('2d std normal, no shape', np.zeros(2), np.ones(2), ()),
      ('2d std normal, int shape', [0, 0], [1, 1], 1),
      ('2d std normal, 1-tuple shape', np.zeros(2), np.ones(2), (1,)),
      ('2d std normal, 2-tuple shape', [0, 0], [1, 1], (2, 2)),
      ('rank 2 std normal, 2-tuple shape', np.zeros(
          (3, 2)), np.ones((3, 2)), (2, 2)),
      ('broadcasted loc', 0, np.ones(3), (2, 2)),
      ('broadcasted scale', np.ones(3), 1, ()),
  ))
  def test_sample_shape(self, mu, sigma, sample_shape, base_dist):
    base = base_dist(mu, sigma)
    bijector = chain.Chain([tfb.Scale(2), tfb.Shift(3), tfb.Tanh()])
    dist = transformed.Transformed(base, bijector)
    def sample_fn(seed, sample_shape):
      return dist.sample(seed=seed, sample_shape=sample_shape)
    samples = self.variant(sample_fn, ignore_argnums=(1,), static_argnums=1)(
        self.seed, sample_shape)

    tfp_bijector = tfb.Chain([tfb.Scale(2), tfb.Shift(3), tfb.Tanh()])
    tfp_dist = tfd.TransformedDistribution(
        conversion.to_tfp(base), tfp_bijector)
    tfp_samples = tfp_dist.sample(sample_shape=sample_shape, seed=self.seed)

    chex.assert_equal_shape([samples, tfp_samples])

  @chex.all_variants
  @parameterized.named_parameters(_with_base_dists(
      ('1d dist, 1d value', 0, 1, 1.),
      ('1d dist, 2d value', 0., 1., np.array([1., 2.])),
      ('2d dist, 1d value', np.zeros(2), np.ones(2), 1.),
      ('2d broadcasted dist, 1d value', np.zeros(2), 1, 1.),
      ('2d dist, 2d value', np.zeros(2), np.ones(2), np.array([1., 2.])),
      ('1d dist, 1d value, edge case', 0, 1, 200.),
  ))
  def test_log_prob(self, mu, sigma, value, base_dist):
    base = base_dist(mu, sigma)
    bijector = chain.Chain([tfb.Scale(2), tfb.Shift(3), tfb.Tanh()])
    dist = transformed.Transformed(base, bijector)
    actual = self.variant(dist.log_prob)(value)

    tfp_bijector = tfb.Chain([tfb.Scale(2), tfb.Shift(3), tfb.Tanh()])
    tfp_dist = tfd.TransformedDistribution(
        conversion.to_tfp(base), tfp_bijector)
    expected = tfp_dist.log_prob(value)
    np.testing.assert_array_equal(actual, expected)

  @chex.all_variants
  @parameterized.named_parameters(_with_base_dists(
      ('1d dist, 1d value', 0, 1, 1.),
      ('1d dist, 2d value', 0., 1., np.array([1., 2.])),
      ('2d dist, 1d value', np.zeros(2), np.ones(2), 1.),
      ('2d broadcasted dist, 1d value', np.zeros(2), 1, 1.),
      ('2d dist, 2d value', np.zeros(2), np.ones(2), np.array([1., 2.])),
      ('1d dist, 1d value, edge case', 0, 1, 200.),
  ))
  def test_prob(self, mu, sigma, value, base_dist):
    base = base_dist(mu, sigma)
    bijector = chain.Chain([tfb.Scale(2), tfb.Shift(3), tfb.Tanh()])
    dist = transformed.Transformed(base, bijector)
    actual = self.variant(dist.prob)(value)

    tfp_bijector = tfb.Chain([tfb.Scale(2), tfb.Shift(3), tfb.Tanh()])
    tfp_dist = tfd.TransformedDistribution(
        conversion.to_tfp(base), tfp_bijector)
    expected = tfp_dist.prob(value)
    np.testing.assert_array_equal(actual, expected)

  @chex.all_variants
  @parameterized.named_parameters(_with_base_dists(
      ('1d std normal, no shape', 0, 1, ()),
      ('1d std normal, int shape', 0, 1, 1),
      ('1d std normal, 1-tuple shape', 0, 1, (1,)),
      ('1d std normal, 2-tuple shape', 0, 1, (2, 2)),
      ('2d std normal, no shape', np.zeros(2), np.ones(2), ()),
      ('2d std normal, int shape', [0, 0], [1, 1], 1),
      ('2d std normal, 1-tuple shape', np.zeros(2), np.ones(2), (1,)),
      ('2d std normal, 2-tuple shape', [0, 0], [1, 1], (2, 2)),
      ('rank 2 std normal, 2-tuple shape', np.zeros(
          (3, 2)), np.ones((3, 2)), (2, 2)),
      ('broadcasted loc', 0, np.ones(3), (2, 2)),
      ('broadcasted scale', np.ones(3), 1, ()),
  ))
  def test_sample_and_log_prob(self, mu, sigma, sample_shape, base_dist):
    base = base_dist(mu, sigma)
    bijector = chain.Chain([tfb.Scale(10), tfb.Tanh(), tfb.Scale(0.1)])
    dist = transformed.Transformed(base, bijector)
    def sample_and_log_prob_fn(seed, sample_shape):
      return dist.sample_and_log_prob(seed=seed, sample_shape=sample_shape)
    samples, log_prob = self.variant(
        sample_and_log_prob_fn, ignore_argnums=(1,), static_argnums=(1,))(
            self.seed, sample_shape)
    expected_samples = bijector.forward(
        base.sample(seed=self.seed, sample_shape=sample_shape))

    tfp_bijector = tfb.Chain([tfb.Scale(10), tfb.Tanh(), tfb.Scale(0.1)])
    tfp_dist = tfd.TransformedDistribution(
        conversion.to_tfp(base), tfp_bijector)
    tfp_samples = tfp_dist.sample(seed=self.seed, sample_shape=sample_shape)
    tfp_log_prob = tfp_dist.log_prob(samples)

    chex.assert_equal_shape([samples, tfp_samples])
    np.testing.assert_allclose(log_prob, tfp_log_prob, rtol=RTOL)
    np.testing.assert_allclose(samples, expected_samples, rtol=RTOL)

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(_with_base_dists(
      ('entropy', 'entropy', 0., 1.),
      ('mean', 'mean', 0, 1),
      ('mean from list params', 'mean', [-1, 1], [1, 2]),
      ('mode', 'mode', 0, 1),
  ))
  def test_method(self, function_string, mu, sigma, base_dist):
    base = base_dist(mu, sigma)
    bijector = chain.Chain([tfb.Identity(), tfb.Scale(2), tfb.Shift(3)])
    dist = transformed.Transformed(base, bijector)

    tfp_bijector = tfb.Chain([tfb.Identity(), tfb.Scale(2), tfb.Shift(3)])
    tfp_dist = tfd.TransformedDistribution(
        conversion.to_tfp(base), tfp_bijector)

    np.testing.assert_allclose(
        self.variant(getattr(dist, function_string))(),
        getattr(tfp_dist, function_string)())

  @chex.all_variants
  @parameterized.named_parameters(
      ('scalar outer, scalar inner, batched input', (), (), (2, 3)),
      ('scalar outer, batched inner, scalar input', (), (2, 3), ()),
      ('scalar outer, batched inner, batched input', (), (2, 3), (2, 3)),
      ('batched outer, scalar inner, scalar input', (2, 3), (), ()),
      ('batched outer, scalar inner, batched input', (2, 3), (), (2, 3)),
      ('batched outer, batched inner, scalar input', (2, 3), (2, 3), ()),
      ('batched outer, batched inner, batched input', (2, 3), (2, 3), (2, 3)),
  )
  def test_batched_bijectors(self, outer_shape, inner_shape, input_shape):
    outer_bij = tfb.Shift(5. * jnp.ones(outer_shape))
    inner_bij = tfb.Shift(7. * jnp.ones(inner_shape))

    dx_bij = chain.Chain([outer_bij, inner_bij])
    tfb_bij = tfb.Chain([outer_bij, inner_bij])

    x = jnp.zeros(input_shape)
    dx_y = self.variant(dx_bij.forward)(x)
    tfb_y = self.variant(tfb_bij.forward)(x)

    chex.assert_equal_shape([dx_y, tfb_y])
    np.testing.assert_allclose(dx_y, tfb_y, rtol=RTOL)

  def test_raises_on_empty_list(self):
    with self.assertRaises(ValueError):
      chain.Chain([])

  def test_raises_on_incompatible_dimensions(self):
    with self.assertRaises(ValueError):
      chain.Chain([jnp.log, block.Block(jnp.exp, 1)])

  @chex.all_variants
  @parameterized.named_parameters(
      ('int16', np.array([1, 2], dtype=np.int16)),
      ('int32', np.array([1, 2], dtype=np.int32)),
      ('int64', np.array([1, 2], dtype=np.int64)),
  )
  def test_integer_inputs(self, inputs):
    bijector = chain.Chain([scalar_affine.ScalarAffine(shift=1.0)])
    output, log_det = self.variant(bijector.forward_and_log_det)(inputs)

    expected_out = jnp.array(inputs, dtype=jnp.float32) + 1.0
    expected_log_det = jnp.zeros_like(inputs, dtype=jnp.float32)

    np.testing.assert_array_equal(output, expected_out)
    np.testing.assert_array_equal(log_det, expected_log_det)

  def test_jittable(self):
    @jax.jit
    def f(x, b):
      return b.forward(x)

    bijector = chain.Chain([scalar_affine.ScalarAffine(0, 1)])
    x = np.zeros(())
    f(x, bijector)

if __name__ == '__main__':
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""A simple mixture of two (possibly heterogeneous) distribution."""

from typing import Tuple

import chex
from distrax._src.distributions import distribution as base_distribution
from distrax._src.utils import conversion
import jax
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp


tfd = tfp.distributions
Array = chex.Array
Numeric = chex.Numeric
PRNGKey = chex.PRNGKey
DistributionLike = base_distribution.DistributionLike
EventT = base_distribution.EventT


class MixtureOfTwo(base_distribution.Distribution):
  """A mixture of two distributions."""

  def __init__(
      self,
      prob_a: Numeric,
      component_a: DistributionLike,
      component_b: DistributionLike):
    """Creates a mixture of two distributions.

    Differently from `MixtureSameFamily` the component distributions are allowed
    to belong to different families.

    Args:
      prob_a: a scalar weight for the `component_a`, is a float or a rank 0
        vector.
      component_a: the first component distribution.
      component_b: the second component distribution.
    """
    super().__init__()
    # Validate inputs.
    chex.assert_rank(prob_a, 0)
    if component_a.event_shape != component_b.event_shape:
      raise ValueError(
          f'The component distributions must have the same event shape, but '
          f'{component_a.event_shape} != {component_b.event_shape}.')
    if component_a.batch_shape != component_b.batch_shape:
      raise ValueError(
          f'The component distributions must have the same batch shape, but '
          f'{component_a.batch_shape} != {component_b.batch_shape}.')
    if component_a.dtype != component_b.dtype:
      raise ValueError(
          'The component distributions must have the same dtype, but'
          f' {component_a.dtype} != {component_b.dtype}.')
    # Store args.
    self._prob_a = prob_a
    self._component_a = conversion.as_distribution(component_a)
    self._component_b = conversion.as_distribution(component_b)

  def _sample_n(self, key: PRNGKey, n: int) -> Array:
    """See `Distribution._sample_n`."""
    key, key_a, key_b, mask_key = jax.random.split(key, num=4)
    mask_from_a = jax.random.bernoulli(mask_key, p=self._prob_a, shape=[n])
    sample_a = self._component_a.sample(seed=key_a, sample_shape=n)
    sample_b = self._component_b.sample(seed=key_b, sample_shape=n)
    mask_from_a = jnp.expand_dims(mask_from_a, tuple(range(1, sample_a.ndim)))
    return jnp.where(mask_from_a, sample_a, sample_b)

  def log_prob(self, value: EventT) -> Array:
    """See `Distribution.log_prob`."""
    logp1 = jnp.log(self._prob_a) + self._component_a.log_prob(value)
    logp2 = jnp.log(1 - self._prob_a) + self._component_b.log_prob(value)
    return jnp.logaddexp(logp1, logp2)

  @property
  def event_shape(self) -> Tuple[int, ...]:
    return self._component_a.event_shape

  @property
  def batch_shape(self) -> Tuple[int, ...]:
    return self._component_a.batch_shape

  @property
  def prob_a(self) -> Numeric:
    return self._prob_a

  @property
  def prob_b(self) -> Numeric:
    return 1. - self._prob_a

  def __getitem__(self, index) -> 'MixtureOfTwo':
    """See `Distribution.__getitem__`."""
    index = base_distribution.to_batch_shape_index(self.batch_shape, index)
    return MixtureOfTwo(
        prob_a=self.prob_a,
        component_a=self._component_a[index],
        component_b=self._component_b[index])



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `monte_carlo.py`."""

from absl.testing import absltest

import chex
from distrax._src.distributions import mvn_diag
from distrax._src.distributions import normal
from distrax._src.utils import monte_carlo
import haiku as hk
import jax
import numpy as np
from tensorflow_probability.substrates import jax as tfp


class McTest(absltest.TestCase):

  def test_estimate_kl_with_dice(self):
    batch_size = 5
    num_actions = 11
    num_samples = 1024
    rng_seq = hk.PRNGSequence(0)

    distribution_a = tfp.distributions.Categorical(
        logits=jax.random.normal(next(rng_seq), [batch_size, num_actions]))
    distribution_b = tfp.distributions.Categorical(
        logits=jax.random.normal(next(rng_seq), [batch_size, num_actions]))

    kl_estim_exact = monte_carlo.estimate_kl_best_effort(
        distribution_a, distribution_b, next(rng_seq), num_samples=num_samples)
    kl_estim_mc = monte_carlo.mc_estimate_kl(
        distribution_a, distribution_b, next(rng_seq), num_samples=num_samples)
    kl = distribution_a.kl_divergence(distribution_b)
    np.testing.assert_allclose(kl, kl_estim_exact, rtol=1e-5)
    np.testing.assert_allclose(kl, kl_estim_mc, rtol=2e-1)

  def test_estimate_continuous_kl_with_dice(self):
    _check_kl_estimator(monte_carlo.mc_estimate_kl, tfp.distributions.Normal)
    _check_kl_estimator(monte_carlo.mc_estimate_kl, normal.Normal)

  def test_estimate_continuous_kl_with_reparameterized(self):
    _check_kl_estimator(monte_carlo.mc_estimate_kl_with_reparameterized,
                        tfp.distributions.Normal)
    _check_kl_estimator(monte_carlo.mc_estimate_kl_with_reparameterized,
                        normal.Normal)

  def test_estimate_mode(self):
    with self.subTest('ScalarEventShape'):
      distribution = normal.Normal(
          loc=np.zeros((4, 5, 100)),
          scale=np.ones((4, 5, 100)))
      # pytype: disable=wrong-arg-types
      mode_estimate = monte_carlo.mc_estimate_mode(
          distribution, rng_key=42, num_samples=100)
      # pytype: enable=wrong-arg-types
      mean_mode_estimate = np.abs(np.mean(mode_estimate))
      self.assertLess(mean_mode_estimate, 1e-3)
    with self.subTest('NonScalarEventShape'):
      distribution = mvn_diag.MultivariateNormalDiag(
          loc=np.zeros((4, 5, 100)),
          scale_diag=np.ones((4, 5, 100)))
      # pytype: disable=wrong-arg-types
      mv_mode_estimate = monte_carlo.mc_estimate_mode(
          distribution, rng_key=42, num_samples=100)
      # pytype: enable=wrong-arg-types
      mean_mv_mode_estimate = np.abs(np.mean(mv_mode_estimate))
      self.assertLess(mean_mv_mode_estimate, 1e-1)
      # The mean of the mode-estimate of the Normal should be a lot closer
      # to 0 compared to the MultivariateNormal, because the 100 less samples
      # are taken and most of the mass in a high-dimensional gaussian is NOT
      # at 0!
      self.assertLess(10 * mean_mode_estimate, mean_mv_mode_estimate)


def _check_kl_estimator(estimator_fn, distribution_fn, num_samples=10000,
                        rtol=1e-1, atol=1e-3, grad_rtol=2e-1, grad_atol=1e-1):
  """Compares the estimator_fn output and gradient to exact KL."""
  rng_key = jax.random.PRNGKey(0)

  def expected_kl(params):
    distribution_a = distribution_fn(**params[0])
    distribution_b = distribution_fn(**params[1])
    return distribution_a.kl_divergence(distribution_b)

  def estimate_kl(params):
    distribution_a = distribution_fn(**params[0])
    distribution_b = distribution_fn(**params[1])
    return estimator_fn(distribution_a, distribution_b, rng_key=rng_key,
                        num_samples=num_samples)

  params = (
      dict(loc=0.0, scale=1.0),
      dict(loc=0.1, scale=1.0),
  )
  expected_value, expected_grad = jax.value_and_grad(expected_kl)(params)
  value, grad = jax.value_and_grad(estimate_kl)(params)

  np.testing.assert_allclose(expected_value, value, rtol=rtol, atol=atol)
  chex.assert_trees_all_close(
      expected_grad, grad, rtol=grad_rtol, atol=grad_atol
  )


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for hmm.py."""

import functools

from absl.testing import absltest
from absl.testing import parameterized
import chex
from distrax._src.distributions import categorical
from distrax._src.distributions import mvn_diag
from distrax._src.distributions import normal
from distrax._src.utils import hmm
import jax
import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions


def _make_models(init_logits, trans_logits, obs_dist_name, obs_params, length):
  """Build distrax HMM and equivalent TFP HMM."""
  obs_dist = {
      "categorical": categorical.Categorical,
      "normal": normal.Normal,
      "mvn_diag": mvn_diag.MultivariateNormalDiag,
  }[obs_dist_name](*obs_params)

  dx_model = hmm.HMM(
      init_dist=categorical.Categorical(init_logits),
      trans_dist=categorical.Categorical(trans_logits),
      obs_dist=obs_dist,
  )

  tfp_obs_dist = {
      "categorical": tfd.Categorical,
      "normal": tfd.Normal,
      "mvn_diag": tfd.MultivariateNormalDiag,
  }[obs_dist_name](*obs_params)

  tfp_model = tfd.HiddenMarkovModel(
      initial_distribution=tfd.Categorical(init_logits),
      transition_distribution=tfd.Categorical(trans_logits),
      observation_distribution=tfp_obs_dist,
      num_steps=length,
  )

  return dx_model, tfp_model


class Function:
  """Overrides lambda __repr__ to "fn" to stabilize test naming across cores."""

  def __init__(self, fn):
    self._fn = fn

  def __call__(self, *args, **kwargs):
    return self._fn(*args, **kwargs)

  def __repr__(self):
    return "fn"


def _test_cases(test_fn):
  return parameterized.product(
      length=(1, 17),
      num_states=(2, 23),
      obs_dist_name_and_params_fn=(
          ("categorical", Function(lambda n: (  # pylint: disable=g-long-lambda
              jax.random.normal(jax.random.PRNGKey(0), (n, 7)),))),
          ("normal", Function(lambda n: (  # pylint: disable=g-long-lambda
              jax.random.normal(jax.random.PRNGKey(0), (n,)),
              jax.random.normal(jax.random.PRNGKey(1), (n,))**2))),
          ("mvn_diag", Function(lambda n: (  # pylint: disable=g-long-lambda
              jax.random.normal(jax.random.PRNGKey(0), (n, 7)),
              jax.random.normal(jax.random.PRNGKey(1), (n, 7))**2))),
      ),
  )(test_fn)


class HMMTest(parameterized.TestCase):

  @chex.all_variants(without_device=False)
  @_test_cases
  def test_sample(self, length, num_states, obs_dist_name_and_params_fn):
    name, params_fn = obs_dist_name_and_params_fn

    logits = jax.random.normal(jax.random.PRNGKey(0), (num_states,))
    matrix = jax.random.normal(jax.random.PRNGKey(1), (num_states, num_states))

    model, tfp_model = _make_models(
        init_logits=logits, trans_logits=matrix, obs_dist_name=name,
        obs_params=params_fn(num_states), length=length)

    states, obs = self.variant(functools.partial(model.sample, seq_len=length))(
        seed=jax.random.PRNGKey(0))
    tfp_obs = tfp_model.sample(seed=jax.random.PRNGKey(0))

    with self.subTest("states"):
      chex.assert_type(states, jnp.int32)
      chex.assert_shape(states, (length,))

    with self.subTest("observations"):
      chex.assert_type(obs, model.obs_dist.dtype)
      chex.assert_shape(obs, (length, *model.obs_dist.event_shape))

    with self.subTest("matches TFP"):
      chex.assert_equal_shape([obs, tfp_obs])

  @chex.all_variants(without_device=False)
  @_test_cases
  def test_forward_backward(
      self, length, num_states, obs_dist_name_and_params_fn):
    name, params_fn = obs_dist_name_and_params_fn

    logits = jax.random.normal(jax.random.PRNGKey(0), (num_states,))
    matrix = jax.random.normal(jax.random.PRNGKey(1), (num_states, num_states))

    model, tfp_model = _make_models(
        init_logits=logits, trans_logits=matrix, obs_dist_name=name,
        obs_params=params_fn(num_states), length=length)

    _, observations = model.sample(seed=jax.random.PRNGKey(42), seq_len=length)
    alphas, betas, marginals, log_prob = self.variant(model.forward_backward)(
        observations)
    tfp_marginal_logits = tfp_model.posterior_marginals(observations).logits
    tfp_marginals = jax.nn.softmax(tfp_marginal_logits)

    with self.subTest("alphas"):
      chex.assert_type(alphas, jnp.float32)
      chex.assert_shape(alphas, (length, num_states))

    with self.subTest("betas"):
      chex.assert_type(betas, jnp.float32)
      chex.assert_shape(betas, (length, num_states))

    with self.subTest("marginals"):
      chex.assert_type(marginals, jnp.float32)
      chex.assert_shape(marginals, (length, num_states))

    with self.subTest("log_prob"):
      chex.assert_type(log_prob, jnp.float32)
      chex.assert_shape(log_prob, ())

    with self.subTest("matches TFP"):
      np.testing.assert_array_almost_equal(marginals, tfp_marginals, decimal=4)

  @chex.all_variants(without_device=False)
  @_test_cases
  def test_viterbi(self, length, num_states, obs_dist_name_and_params_fn):
    name, params_fn = obs_dist_name_and_params_fn

    logits = jax.random.normal(jax.random.PRNGKey(0), (num_states,))
    matrix = jax.random.normal(jax.random.PRNGKey(1), (num_states, num_states))

    model, tfp_model = _make_models(
        init_logits=logits, trans_logits=matrix, obs_dist_name=name,
        obs_params=params_fn(num_states), length=length)

    _, observations = model.sample(seed=jax.random.PRNGKey(42), seq_len=length)
    most_likely_states = self.variant(model.viterbi)(observations)
    tfp_mode = tfp_model.posterior_mode(observations)

    with self.subTest("shape"):
      chex.assert_shape(most_likely_states, (length,))

    with self.subTest("matches TFP"):
      np.testing.assert_array_equal(most_likely_states, tfp_mode)

  @chex.all_variants(without_device=False)
  def test_viterbi_matches_specific_example(self):
    loc = jnp.array([0.0, 1.0, 2.0, 3.0])
    scale = jnp.array(0.25)
    initial = jnp.array([0.25, 0.25, 0.25, 0.25])
    trans = jnp.array([[0.9, 0.1, 0.0, 0.0],
                       [0.1, 0.8, 0.1, 0.0],
                       [0.0, 0.1, 0.8, 0.1],
                       [0.0, 0.0, 0.1, 0.9]])

    observations = jnp.array([0.1, 0.2, 0.3, 0.4, 0.5, 3.0, 2.9, 2.8, 2.7, 2.6])

    model = hmm.HMM(
        init_dist=categorical.Categorical(probs=initial),
        trans_dist=categorical.Categorical(probs=trans),
        obs_dist=normal.Normal(loc, scale))

    inferred_states = self.variant(model.viterbi)(observations)
    expected_states = [0, 0, 0, 0, 1, 2, 3, 3, 3, 3]
    np.testing.assert_array_equal(inferred_states, expected_states)


if __name__ == "__main__":
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `mvn_from_bijector.py`."""

from typing import Tuple

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.bijectors import linear
from distrax._src.bijectors.diag_linear import DiagLinear
from distrax._src.bijectors.triangular_linear import TriangularLinear
from distrax._src.distributions.mvn_from_bijector import MultivariateNormalFromBijector

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions

Array = chex.Array


class MockLinear(linear.Linear):
  """A mock linear bijector."""

  def __init__(self, event_dims: int):
    # pytype: disable=wrong-arg-types  # jnp-type
    super().__init__(event_dims, batch_shape=(), dtype=float)
    # pytype: enable=wrong-arg-types

  def forward_and_log_det(self, x: Array) -> Tuple[Array, Array]:
    """Computes y = f(x) and log|det J(f)(x)|."""
    return x, jnp.zeros_like(x)[:-1]


class MultivariateNormalFromBijectorTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('loc is 0d', 4, np.zeros(shape=())),
      ('loc and scale dims not compatible', 3, np.zeros((4,))),
  )
  def test_raises_on_wrong_inputs(self, event_dims, loc):
    bij = MockLinear(event_dims)
    with self.assertRaises(ValueError):
      MultivariateNormalFromBijector(loc, bij)

  @parameterized.named_parameters(
      ('no broadcast', np.ones((4,)), np.zeros((4,)), (4,)),
      ('broadcasted loc', np.ones((3, 4)), np.zeros((4,)), (3, 4)),
      ('broadcasted diag', np.ones((4,)), np.zeros((3, 4)), (3, 4)),
  )
  def test_loc_scale_and_shapes(self, diag, loc, expected_shape):
    scale = DiagLinear(diag)
    batch_shape = jnp.broadcast_shapes(diag.shape, loc.shape)[:-1]
    dist = MultivariateNormalFromBijector(loc, scale)
    np.testing.assert_allclose(dist.loc, np.zeros(expected_shape))
    self.assertTrue(scale.same_as(dist.scale))
    self.assertEqual(dist.event_shape, (4,))
    self.assertEqual(dist.batch_shape, batch_shape)

  @chex.all_variants
  def test_sample(self):
    prng = hk.PRNGSequence(jax.random.PRNGKey(42))
    diag = 0.5 + jax.random.uniform(next(prng), (4,))
    loc = jax.random.normal(next(prng), (4,))
    scale = DiagLinear(diag)
    dist = MultivariateNormalFromBijector(loc, scale)
    num_samples = 100_000
    sample_fn = lambda seed: dist.sample(seed=seed, sample_shape=num_samples)
    samples = self.variant(sample_fn)(jax.random.PRNGKey(2000))
    self.assertEqual(samples.shape, (num_samples, 4))
    np.testing.assert_allclose(jnp.mean(samples, axis=0), loc, rtol=0.1)
    np.testing.assert_allclose(jnp.std(samples, axis=0), diag, rtol=0.1)

  @chex.all_variants
  def test_log_prob(self):
    prng = hk.PRNGSequence(jax.random.PRNGKey(42))
    diag = 0.5 + jax.random.uniform(next(prng), (4,))
    loc = jax.random.normal(next(prng), (4,))
    scale = DiagLinear(diag)
    dist = MultivariateNormalFromBijector(loc, scale)
    values = jax.random.normal(next(prng), (5, 4))
    tfp_dist = tfd.MultivariateNormalDiag(loc=loc, scale_diag=diag)
    np.testing.assert_allclose(
        self.variant(dist.log_prob)(values), tfp_dist.log_prob(values),
        rtol=2e-7)

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('no broadcast', (4,), (4,)),
      ('broadcasted loc', (3, 4), (4,)),
      ('broadcasted diag', (4,), (3, 4)),
  )
  def test_mean_median_mode(self, diag_shape, loc_shape):
    prng = hk.PRNGSequence(jax.random.PRNGKey(42))
    diag = jax.random.normal(next(prng), diag_shape)
    loc = jax.random.normal(next(prng), loc_shape)
    scale = DiagLinear(diag)
    batch_shape = jnp.broadcast_shapes(diag_shape, loc_shape)[:-1]
    dist = MultivariateNormalFromBijector(loc, scale)
    for method in ['mean', 'median', 'mode']:
      with self.subTest(method=method):
        fn = self.variant(getattr(dist, method))
        np.testing.assert_allclose(
            fn(), jnp.broadcast_to(loc, batch_shape + loc.shape[-1:]))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('no broadcast', (4,), (4,)),
      ('broadcasted loc', (3, 4), (4,)),
      ('broadcasted diag', (4,), (3, 4)),
  )
  def test_variance_stddev_covariance_diag(self, scale_shape, loc_shape):
    prng = hk.PRNGSequence(jax.random.PRNGKey(42))
    scale_diag = jax.random.normal(next(prng), scale_shape)
    loc = jax.random.normal(next(prng), loc_shape)
    scale = DiagLinear(scale_diag)
    batch_shape = jnp.broadcast_shapes(scale_shape[:-1], loc_shape[:-1])
    dist = MultivariateNormalFromBijector(loc, scale)
    for method in ['variance', 'stddev', 'covariance']:
      with self.subTest(method=method):
        fn = self.variant(getattr(dist, method))
        if method == 'variance':
          expected_result = jnp.broadcast_to(
              jnp.square(scale_diag), batch_shape + loc.shape[-1:])
        elif method == 'stddev':
          expected_result = jnp.broadcast_to(
              jnp.abs(scale_diag), batch_shape + loc.shape[-1:])
        elif method == 'covariance':
          expected_result = jnp.broadcast_to(
              jnp.vectorize(jnp.diag, signature='(k)->(k,k)')(
                  jnp.square(scale_diag)),
              batch_shape + loc.shape[-1:] + loc.shape[-1:])
        else:
          raise ValueError(f'Unsupported method: {method}')
        np.testing.assert_allclose(fn(), expected_result, rtol=5e-3)

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('no broadcast', (4, 4), (4,)),
      ('broadcasted loc', (3, 4, 4), (4,)),
      ('broadcasted diag', (4, 4), (3, 4)),
  )
  def test_variance_stddev_covariance_no_diag(self, scale_shape, loc_shape):
    prng = hk.PRNGSequence(jax.random.PRNGKey(42))
    scale_tril = jnp.tril(jax.random.normal(next(prng), scale_shape))
    loc = jax.random.normal(next(prng), loc_shape)
    scale = TriangularLinear(matrix=scale_tril, is_lower=True)
    batch_shape = jnp.broadcast_shapes(scale_shape[:-2], loc_shape[:-1])
    dist = MultivariateNormalFromBijector(loc, scale)
    for method in ['variance', 'stddev', 'covariance']:
      with self.subTest(method=method):
        fn = self.variant(getattr(dist, method))
        scale_tril_t = jnp.vectorize(
            jnp.transpose, signature='(k,k)->(k,k)')(scale_tril)
        scale_times_scale_t = jnp.matmul(scale_tril, scale_tril_t)
        if method == 'variance':
          expected_result = jnp.vectorize(jnp.diag, signature='(k,k)->(k)')(
              scale_times_scale_t)
          expected_result = jnp.broadcast_to(
              expected_result, batch_shape + loc.shape[-1:])
        elif method == 'stddev':
          expected_result = jnp.vectorize(jnp.diag, signature='(k,k)->(k)')(
              jnp.sqrt(scale_times_scale_t))
          expected_result = jnp.broadcast_to(
              expected_result, batch_shape + loc.shape[-1:])
        elif method == 'covariance':
          expected_result = jnp.broadcast_to(
              scale_times_scale_t, batch_shape + scale_tril.shape[-2:])
        np.testing.assert_allclose(fn(), expected_result, rtol=5e-3)

  @chex.all_variants(with_pmap=False)
  def test_kl_divergence_diag_distributions(self):
    prng = hk.PRNGSequence(jax.random.PRNGKey(42))

    scale_diag1 = 0.1 + jax.random.uniform(next(prng), (3, 4))
    loc1 = jax.random.normal(next(prng), (1, 4))
    dist1_distrax = MultivariateNormalFromBijector(
        loc=loc1,
        scale=DiagLinear(scale_diag1),
    )
    dist1_tfp = tfd.MultivariateNormalDiag(
        loc=loc1, scale_diag=scale_diag1)

    scale_diag2 = 0.1 + jax.random.uniform(next(prng), (4,))
    loc2 = jax.random.normal(next(prng), (4,))
    dist2_distrax = MultivariateNormalFromBijector(
        loc=loc2,
        scale=DiagLinear(scale_diag2),
    )
    dist2_tfp = tfd.MultivariateNormalDiag(
        loc=loc2, scale_diag=scale_diag2)

    expected_result1 = dist1_tfp.kl_divergence(dist2_tfp)
    expected_result2 = dist2_tfp.kl_divergence(dist1_tfp)

    for mode in ['distrax_to_distrax', 'distrax_to_tfp', 'tfp_to_distrax']:
      with self.subTest(mode=mode):
        if mode == 'distrax_to_distrax':
          result1 = self.variant(dist1_distrax.kl_divergence)(dist2_distrax)
          result2 = self.variant(dist2_distrax.kl_divergence)(dist1_distrax)
        elif mode == 'distrax_to_tfp':
          result1 = self.variant(dist1_distrax.kl_divergence)(dist2_tfp)
          result2 = self.variant(dist2_distrax.kl_divergence)(dist1_tfp)
        elif mode == 'tfp_to_distrax':
          result1 = self.variant(dist1_tfp.kl_divergence)(dist2_distrax)
          result2 = self.variant(dist2_tfp.kl_divergence)(dist1_distrax)
        else:
          raise ValueError(f'Unsupported mode: {mode}')
        np.testing.assert_allclose(result1, expected_result1, rtol=1e-3)
        np.testing.assert_allclose(result2, expected_result2, rtol=1e-3)

  @chex.all_variants(with_pmap=False)
  def test_kl_divergence_non_diag_distributions(self):
    prng = hk.PRNGSequence(jax.random.PRNGKey(42))

    scale_tril1 = jnp.tril(jax.random.normal(next(prng), (3, 4, 4)))
    loc1 = jax.random.normal(next(prng), (1, 4))
    dist1_distrax = MultivariateNormalFromBijector(
        loc=loc1,
        scale=TriangularLinear(matrix=scale_tril1),
    )
    dist1_tfp = tfd.MultivariateNormalTriL(loc=loc1, scale_tril=scale_tril1)

    scale_tril2 = jnp.tril(jax.random.normal(next(prng), (4, 4)))
    loc2 = jax.random.normal(next(prng), (4,))
    dist2_distrax = MultivariateNormalFromBijector(
        loc=loc2,
        scale=TriangularLinear(matrix=scale_tril2),
    )
    dist2_tfp = tfd.MultivariateNormalTriL(loc=loc2, scale_tril=scale_tril2)

    expected_result1 = dist1_tfp.kl_divergence(dist2_tfp)
    expected_result2 = dist2_tfp.kl_divergence(dist1_tfp)

    for mode in ['distrax_to_distrax', 'distrax_to_tfp', 'tfp_to_distrax']:
      with self.subTest(mode=mode):
        if mode == 'distrax_to_distrax':
          result1 = self.variant(dist1_distrax.kl_divergence)(dist2_distrax)
          result2 = self.variant(dist2_distrax.kl_divergence)(dist1_distrax)
        elif mode == 'distrax_to_tfp':
          result1 = self.variant(dist1_distrax.kl_divergence)(dist2_tfp)
          result2 = self.variant(dist2_distrax.kl_divergence)(dist1_tfp)
        elif mode == 'tfp_to_distrax':
          result1 = self.variant(dist1_tfp.kl_divergence)(dist2_distrax)
          result2 = self.variant(dist2_tfp.kl_divergence)(dist1_distrax)
        else:
          raise ValueError(f'Unsupported mode: {mode}')
        np.testing.assert_allclose(result1, expected_result1, rtol=1e-3)
        np.testing.assert_allclose(result2, expected_result2, rtol=1e-3)

  def test_kl_divergence_raises_on_incompatible_distributions(self):
    dim = 4
    dist1 = MultivariateNormalFromBijector(
        loc=jnp.zeros((dim,)),
        scale=DiagLinear(diag=jnp.ones((dim,))),
    )
    dim = 5
    dist2 = MultivariateNormalFromBijector(
        loc=jnp.zeros((dim,)),
        scale=DiagLinear(diag=jnp.ones((dim,))),
    )
    with self.assertRaises(ValueError):
      dist1.kl_divergence(dist2)


if __name__ == '__main__':
  jax.config.update('jax_threefry_partitionable', False)
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `multinomial.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions import multinomial
from distrax._src.utils import equivalence
from distrax._src.utils import math
import jax
import jax.experimental
import jax.numpy as jnp
import numpy as np
from scipy import stats


class MultinomialTest(equivalence.EquivalenceTest):

  def setUp(self):
    super().setUp()
    self._init_distr_cls(multinomial.Multinomial)
    self.total_count = np.asarray(
        [4, 3], dtype=np.float32)  # float dtype required for TFP
    self.probs = 0.5 * np.asarray([0.1, 0.4, 0.2, 0.3])  # unnormalized
    self.logits = np.log(self.probs)

  @parameterized.named_parameters(
      ('from probs', False),
      ('from logits', True))
  def test_parameters(self, from_logits):
    if from_logits:
      dist_params = {'logits': self.logits, 'total_count': self.total_count}
    else:
      dist_params = {'probs': self.probs, 'total_count': self.total_count}
    dist = self.distrax_cls(**dist_params)
    self.assertion_fn(atol=1e-6, rtol=1e-3)(
        dist.logits, np.tile(math.normalize(logits=self.logits), (2, 1)))
    self.assertion_fn(atol=1e-6, rtol=1e-3)(
        dist.probs, np.tile(math.normalize(probs=self.probs), (2, 1)))

  @parameterized.named_parameters(
      ('probs and logits', {
          'total_count': 3, 'logits': [0.1, -0.2], 'probs': [0.6, 0.4]}),
      ('both probs and logits are None', {
          'total_count': 3, 'logits': None, 'probs': None}),
      ('logits are 0d', {'total_count': 3, 'logits': 3.}),
      ('probs are 0d', {'total_count': 3, 'probs': 1.}),
      ('logits have wrong dim', {'total_count': 3, 'logits': np.ones((4, 1))}),
      ('probs have wrong dim', {'total_count': 3, 'probs': np.ones((4, 1))}),
      ('bool dtype', {
          'total_count': 3, 'logits': [0.1, 0.], 'dtype': jnp.bool_}),
      ('complex64 dtype', {
          'total_count': 3, 'logits': [0.1, 0.], 'dtype': jnp.complex64}),
      ('complex128 dtype', {
          'total_count': 3, 'logits': [0.1, 0.], 'dtype': jnp.complex128}),
  )
  def test_raises_on_invalid_inputs(self, dist_params):
    with self.assertRaises(ValueError):
      self.distrax_cls(**dist_params)

  @parameterized.named_parameters(
      ('1d logits', {'logits': [0.0, 1.0, -0.5]}),
      ('1d probs', {'probs': [0.2, 0.5, 0.3]}),
      ('2d logits', {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]}),
      ('2d probs', {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]}),
  )
  def test_event_shape(self, dist_params):
    dist_params = {k: jnp.asarray(v) for k, v in dist_params.items()}
    dist_params.update({'total_count': self.total_count})
    super()._test_event_shape((), dist_params)

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d logits, no shape',
       {'logits': [0.0, 1.0, -0.5]},
       ()),
      ('1d probs, no shape',
       {'probs': [0.2, 0.5, 0.3]},
       ()),
      ('1d logits, int shape',
       {'logits': [0.0, 1.0, -0.5]},
       1),
      ('1d probs, int shape',
       {'probs': [0.2, 0.5, 0.3]},
       1),
      ('1d logits, 1-tuple shape',
       {'logits': [0.0, 1.0, -0.5]},
       (1,)),
      ('1d probs, 1-tuple shape',
       {'probs': [0.2, 0.5, 0.3]},
       (1,)),
      ('1d logits, 2-tuple shape',
       {'logits': [0.0, 1.0, -0.5]},
       (5, 4)),
      ('1d probs, 2-tuple shape',
       {'probs': [0.2, 0.5, 0.3]},
       (5, 4)),
      ('2d logits, no shape',
       {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]},
       ()),
      ('2d probs, no shape',
       {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]},
       ()),
      ('2d logits, int shape',
       {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]},
       4),
      ('2d probs, int shape',
       {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]},
       4),
      ('2d logits, 1-tuple shape',
       {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]},
       (5,)),
      ('2d probs, 1-tuple shape',
       {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]},
       (5,)),
      ('2d logits, 2-tuple shape',
       {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]},
       (5, 4)),
      ('2d probs, 2-tuple shape',
       {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]},
       (5, 4)),
  )
  def test_sample_shape_with_int_total_count(
      self, dist_params, sample_shape):
    dist_params = {k: jnp.asarray(v) for k, v in dist_params.items()}
    dist_params.update({
        'total_count': 3,
    })
    super()._test_sample_shape(
        dist_args=(),
        dist_kwargs=dist_params,
        sample_shape=sample_shape)

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d logits, no shape',
       {'logits': [0.0, 1.0, -0.5]},
       ()),
      ('1d probs, no shape',
       {'probs': [0.2, 0.5, 0.3]},
       ()),
      ('1d logits, int shape',
       {'logits': [0.0, 1.0, -0.5]},
       1),
      ('1d probs, int shape',
       {'probs': [0.2, 0.5, 0.3]},
       1),
      ('1d logits, 1-tuple shape',
       {'logits': [0.0, 1.0, -0.5]},
       (1,)),
      ('1d probs, 1-tuple shape',
       {'probs': [0.2, 0.5, 0.3]},
       (1,)),
      ('1d logits, 2-tuple shape',
       {'logits': [0.0, 1.0, -0.5]},
       (5, 4)),
      ('1d probs, 2-tuple shape',
       {'probs': [0.2, 0.5, 0.3]},
       (5, 4)),
      ('2d logits, no shape',
       {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]},
       ()),
      ('2d probs, no shape',
       {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]},
       ()),
      ('2d logits, int shape',
       {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]},
       4),
      ('2d probs, int shape',
       {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]},
       4),
      ('2d logits, 1-tuple shape',
       {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]},
       (5,)),
      ('2d probs, 1-tuple shape',
       {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]},
       (5,)),
      ('2d logits, 2-tuple shape',
       {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]},
       (5, 4)),
      ('2d probs, 2-tuple shape',
       {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]},
       (5, 4)),
  )
  def test_sample_shape_with_1d_total_count(
      self, dist_params, sample_shape):
    dist_params = {k: jnp.asarray(v) for k, v in dist_params.items()}
    dist_params.update({
        'total_count': np.asarray([4, 3], dtype=np.float32),
    })
    super()._test_sample_shape(
        dist_args=(),
        dist_kwargs=dist_params,
        sample_shape=sample_shape)

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d logits, no shape',
       {'logits': [0.0, 1.0, -0.5]},
       ()),
      ('1d probs, no shape',
       {'probs': [0.2, 0.5, 0.3]},
       ()),
      ('1d logits, int shape',
       {'logits': [0.0, 1.0, -0.5]},
       1),
      ('1d probs, int shape',
       {'probs': [0.2, 0.5, 0.3]},
       1),
      ('1d logits, 1-tuple shape',
       {'logits': [0.0, 1.0, -0.5]},
       (1,)),
      ('1d probs, 1-tuple shape',
       {'probs': [0.2, 0.5, 0.3]},
       (1,)),
      ('1d logits, 2-tuple shape',
       {'logits': [0.0, 1.0, -0.5]},
       (5, 4)),
      ('1d probs, 2-tuple shape',
       {'probs': [0.2, 0.5, 0.3]},
       (5, 4)),
      ('2d logits, no shape',
       {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]},
       ()),
      ('2d probs, no shape',
       {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]},
       ()),
      ('2d logits, int shape',
       {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]},
       4),
      ('2d probs, int shape',
       {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]},
       4),
      ('2d logits, 1-tuple shape',
       {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]},
       (5,)),
      ('2d probs, 1-tuple shape',
       {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]},
       (5,)),
      ('2d logits, 2-tuple shape',
       {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]},
       (5, 4)),
      ('2d probs, 2-tuple shape',
       {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]},
       (5, 4)),
  )
  def test_sample_shape_with_2d_total_count(
      self, dist_params, sample_shape):
    dist_params = {k: jnp.asarray(v) for k, v in dist_params.items()}
    total_count = np.asarray(
        [[4, 3], [5, 4], [3, 2], [1, 4]], dtype=np.float32)
    dist_params.update({'total_count': total_count})
    super()._test_sample_shape(
        dist_args=(),
        dist_kwargs=dist_params,
        sample_shape=sample_shape)

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d logits, no shape',
       {'logits': [0.0, 1.0, -0.5]},
       ()),
      ('1d probs, no shape',
       {'probs': [0.2, 0.5, 0.3]},
       ()),
      ('1d logits, int shape',
       {'logits': [0.0, 1.0, -0.5]},
       1),
      ('1d probs, int shape',
       {'probs': [0.2, 0.5, 0.3]},
       1),
      ('1d logits, 1-tuple shape',
       {'logits': [0.0, 1.0, -0.5]},
       (1,)),
      ('1d probs, 1-tuple shape',
       {'probs': [0.2, 0.5, 0.3]},
       (1,)),
      ('1d logits, 2-tuple shape',
       {'logits': [0.0, 1.0, -0.5]},
       (5, 4)),
      ('1d probs, 2-tuple shape',
       {'probs': [0.2, 0.5, 0.3]},
       (5, 4)),
      ('2d logits, no shape',
       {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]},
       ()),
      ('2d probs, no shape',
       {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]},
       ()),
      ('2d logits, int shape',
       {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]},
       4),
      ('2d probs, int shape',
       {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]},
       4),
      ('2d logits, 1-tuple shape',
       {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]},
       (5,)),
      ('2d probs, 1-tuple shape',
       {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]},
       (5,)),
      ('2d logits, 2-tuple shape',
       {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]},
       (5, 4)),
      ('2d probs, 2-tuple shape',
       {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]},
       (5, 4)),
  )
  def test_sum_samples(
      self, dist_params, sample_shape):
    dist_params = {k: jnp.asarray(v) for k, v in dist_params.items()}
    total_count = np.asarray(
        [[4, 3], [5, 4], [3, 2], [1, 4]], dtype=np.float32)
    dist_params.update({'total_count': total_count})
    dist = self.distrax_cls(**dist_params)
    sample_fn = self.variant(
        lambda key: dist.sample(seed=key, sample_shape=sample_shape))
    samples = sample_fn(self.key)
    sum_samples = jnp.sum(samples, axis=-1)
    self.assertion_fn(atol=1e-6, rtol=1e-3)(
        np.asarray(sum_samples, dtype=np.float32),
        np.broadcast_to(total_count, sum_samples.shape))

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d logits, no shape',
       {'logits': [0.0, 1.0, -0.5]},
       ()),
      ('1d probs, no shape',
       {'probs': [0.2, 0.5, 0.3]},
       ()),
      ('1d logits, int shape',
       {'logits': [0.0, 1.0, -0.5]},
       1),
      ('1d probs, int shape',
       {'probs': [0.2, 0.5, 0.3]},
       1),
      ('1d logits, 1-tuple shape',
       {'logits': [0.0, 1.0, -0.5]},
       (1,)),
      ('1d probs, 1-tuple shape',
       {'probs': [0.2, 0.5, 0.3]},
       (1,)),
      ('1d logits, 2-tuple shape',
       {'logits': [0.0, 1.0, -0.5]},
       (5, 4)),
      ('1d probs, 2-tuple shape',
       {'probs': [0.2, 0.5, 0.3]},
       (5, 4)),
      ('2d logits, no shape',
       {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]},
       ()),
      ('2d probs, no shape',
       {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]},
       ()),
      ('2d logits, int shape',
       {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]},
       4),
      ('2d probs, int shape',
       {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]},
       4),
      ('2d logits, 1-tuple shape',
       {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]},
       (5,)),
      ('2d probs, 1-tuple shape',
       {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]},
       (5,)),
      ('2d logits, 2-tuple shape',
       {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]},
       (5, 4)),
      ('2d probs, 2-tuple shape',
       {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]},
       (5, 4)),
  )
  def test_sample_and_log_prob(self, dist_params, sample_shape):
    dist_params = {k: jnp.asarray(v) for k, v in dist_params.items()}
    total_count = np.asarray(
        [[4, 3], [5, 4], [3, 2], [1, 4]], dtype=np.float32)
    dist_params.update({'total_count': total_count})
    super()._test_sample_and_log_prob(
        dist_args=(),
        dist_kwargs=dist_params,
        sample_shape=sample_shape,
        assertion_fn=self.assertion_fn(atol=1e-6, rtol=1e-3))

  @chex.all_variants
  @parameterized.named_parameters(
      ('int32', jnp.int32),
      ('int64', jnp.int64),
      ('uint32', jnp.uint32),
      ('uint64', jnp.uint64),
      ('float32', jnp.float32),
      ('float64', jnp.float64))
  def test_sample_dtype(self, dtype):
    with jax.experimental.enable_x64(dtype.dtype.itemsize == 8):
      dist_params = {
          'logits': self.logits,
          'dtype': dtype,
          'total_count': self.total_count,
      }
      dist = self.distrax_cls(**dist_params)
      samples = self.variant(dist.sample)(seed=self.key)
      self.assertEqual(samples.dtype, dist.dtype)
      chex.assert_type(samples, dtype)

  @chex.all_variants
  def test_sample_extreme_probs(self):
    dist_params = {
        'probs': np.asarray([1., 0., 0., 0.]), 'total_count': 10}
    dist = self.distrax_cls(**dist_params)
    sample_fn = self.variant(
        lambda key: dist.sample(seed=key, sample_shape=100))
    samples = sample_fn(self.key)
    np.testing.assert_equal(np.unique(samples[..., 0]), 10)
    np.testing.assert_equal(np.unique(samples[..., 1:]), 0)

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d logits, 1 input',
       {'logits': [0.0, 0.5, -0.5]},
       [2, 1, 0]),
      ('1d logits, 2 inputs',
       {'logits': [0.0, 0.5, -0.5]},
       [[1, 2, 0], [0, 1, 2]]),
      ('2d logits, 2 inputs',
       {'logits': [[0.0, 0.5, -0.5], [-0.1, 0.1, 0.1]]},
       [[1, 0, 2], [1, 1, 1]]),
      ('2d logits, rank-3 inputs',
       {'logits': [[0.0, 0.5, -0.5], [-0.1, 0.1, 0.1]]},
       np.asarray([[1, 2, 0], [1, 0, 2]])[None, ...]),
      ('1d probs, 1 input',
       {'probs': [0.3, 0.2, 0.5]},
       [1, 2, 0]),
      ('1d probs, 2 inputs',
       {'probs': [0.3, 0.2, 0.5]},
       [[1, 0, 2], [1, 1, 1]]),
      ('2d probs, 2 inputs',
       {'probs': [[0.2, 0.4, 0.4], [0.1, 0.2, 0.7]]},
       [[1, 2, 0], [2, 1, 0]]),
      ('2d probs, rank-3 inputs',
       {'probs': [[0.2, 0.4, 0.4], [0.1, 0.2, 0.7]]},
       np.asarray([[1, 0, 2], [1, 1, 1]])[None, ...]),
  )
  def test_log_prob(self, dist_params, value):
    dist_params = {k: jnp.asarray(v) for k, v in dist_params.items()}
    dist_params.update({'total_count': 3})
    value = jnp.asarray(value)
    super()._test_attribute(
        attribute_string='log_prob',
        dist_kwargs=dist_params,
        call_args=(value,),
        assertion_fn=self.assertion_fn(atol=1e-6, rtol=1e-3))

  @chex.all_variants(with_jit=False, with_pmap=False)
  def test_log_prob_extreme_probs(self):
    dist_params = {
        'probs': np.array([0.0, 1.0, 0.0]),
        'total_count': 3,
    }
    value = np.array([[0, 3, 0], [1, 1, 1]])
    expected_result = np.asarray([0., -np.inf])
    dist = self.distrax_cls(**dist_params)
    np.testing.assert_allclose(
        self.variant(dist.log_prob)(value), expected_result, atol=1e-5)

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('from 2d logits',
       {'logits': np.asarray([[0.0, 0.5, -0.5], [-0.2, 0.3, 0.5]])}),
      ('from 2d probs',
       {'probs': np.asarray([[0.1, 0.5, 0.4], [0.2, 0.4, 0.4]])}),
  )
  def test_entropy(self, dist_params):
    # The TFP Multinomial does not implement `entropy`, so we use scipy for
    # the tests.
    dist_params.update({
        'total_count': np.asarray([3, 10]),
    })
    dist = self.distrax_cls(**dist_params)
    entropy = []
    for probs, counts in zip(dist.probs, dist.total_count):
      entropy.append(stats.multinomial(n=counts, p=probs).entropy())
    self.assertion_fn(atol=1e-6, rtol=1e-3)(
        self.variant(dist.entropy)(), np.asarray(entropy))

  @chex.all_variants(with_pmap=False)
  def test_entropy_extreme_probs(self):
    dist_params = {
        'probs': np.asarray([1.0, 0.0, 0.0]),
        'total_count': np.asarray([3, 10]),
    }
    dist = self.distrax_cls(**dist_params)
    expected_result = np.asarray([0., 0.])
    np.testing.assert_allclose(
        self.variant(dist.entropy)(), expected_result, atol=3e-4)

  @chex.all_variants(with_pmap=False)
  def test_entropy_scalar(self):
    # The TFP Multinomial does not implement `entropy`, so we use scipy for
    # the tests.
    probs = np.asarray([0.1, 0.5, 0.4])
    total_count = 5
    scipy_entropy = stats.multinomial(n=total_count, p=probs).entropy()
    distrax_entropy_fn = self.variant(
        lambda x, y: multinomial.Multinomial._entropy_scalar(total_count, x, y))
    self.assertion_fn(atol=1e-6, rtol=1e-3)(
        distrax_entropy_fn(probs, np.log(probs)), scipy_entropy)

  @chex.all_variants(with_pmap=False)
  def test_entropy_scalar_extreme_probs(self):
    probs = np.asarray([1., 0., 0.])
    total_count = 5
    expected_result = 0.
    distrax_entropy_fn = self.variant(
        lambda x, y: multinomial.Multinomial._entropy_scalar(total_count, x, y))
    np.testing.assert_allclose(
        distrax_entropy_fn(probs, np.log(probs)), expected_result, atol=1e-5)

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('variance; from 2d logits',
       'variance', {'logits': [[0.0, 0.5, -0.5], [-0.2, 0.3, 0.5]]}),
      ('variance; from 2d probs',
       'variance', {'probs': [[0.1, 0.5, 0.4], [0.2, 0.4, 0.4]]}),
      ('mean; from 2d logits',
       'mean', {'logits': [[0.0, 0.5, -0.5], [-0.2, 0.3, 0.5]]}),
      ('mean; from 2d probs',
       'mean', {'probs': [[0.1, 0.5, 0.4], [0.2, 0.4, 0.4]]}),
      ('covariance; from 2d logits',
       'covariance', {'logits': [[0.0, 0.5, -0.5], [-0.2, 0.3, 0.5]]}),
      ('covariance; from 2d probs',
       'covariance', {'probs': [[0.1, 0.5, 0.4], [0.2, 0.4, 0.4]]}),
  )
  def test_method(self, function_string, dist_params):
    dist_params = {k: jnp.asarray(v) for k, v in dist_params.items()}
    total_count = np.asarray(
        [[4, 3], [5, 4], [3, 2], [1, 4]], dtype=np.float32)
    dist_params.update({'total_count': total_count})
    super()._test_attribute(
        attribute_string=function_string,
        dist_kwargs=dist_params,
        assertion_fn=self.assertion_fn(atol=1e-6, rtol=1e-3))

  def test_jittable(self):
    super()._test_jittable(
        dist_kwargs={
            'probs': np.asarray([1.0, 0.0, 0.0]),
            'total_count': np.asarray([3, 10])
        },
        assertion_fn=self.assertion_fn(atol=3e-4, rtol=1e-3))

  @parameterized.named_parameters(
      ('single element', 2),
      ('range', slice(-1)),
      ('range_2', (slice(None), slice(-1))),
  )
  def test_slice(self, slice_):
    logits = jnp.array(np.random.randn(3, 4, 5))
    probs = jax.nn.softmax(jnp.array(np.random.randn(3, 4, 5)), axis=-1)
    total_count = jnp.full((3, 4), fill_value=2)
    dist1 = self.distrax_cls(total_count=total_count, logits=logits)
    dist2 = self.distrax_cls(total_count=total_count, probs=probs)
    self.assertion_fn(atol=1e-6, rtol=1e-3)(
        dist2[slice_].total_count, total_count[slice_])
    self.assertion_fn(atol=1e-6, rtol=1e-3)(
        jax.nn.softmax(dist1[slice_].logits, axis=-1),
        jax.nn.softmax(logits[slice_], axis=-1))
    self.assertion_fn(atol=1e-6, rtol=1e-3)(dist2[slice_].probs, probs[slice_])

  def test_slice_ellipsis(self):
    logits = jnp.array(np.random.randn(4, 4, 5))
    probs = jax.nn.softmax(jnp.array(np.random.randn(4, 4, 5)), axis=-1)
    total_count_value = 2
    total_count = jnp.full((4, 4), fill_value=total_count_value)
    dist1 = self.distrax_cls(total_count=total_count_value, logits=logits)
    dist2 = self.distrax_cls(total_count=total_count_value, probs=probs)
    self.assertion_fn(atol=1e-6, rtol=1e-3)(
        dist1[..., -1].total_count, total_count[..., -1])
    self.assertion_fn(atol=1e-6, rtol=1e-3)(
        dist2[..., -1].total_count, total_count[..., -1])
    self.assertion_fn(atol=1e-6, rtol=1e-3)(
        jax.nn.softmax(dist1[..., -1].logits, axis=-1),
        jax.nn.softmax(logits[:, -1], axis=-1))
    self.assertion_fn(atol=1e-6, rtol=1e-3)(dist2[..., -1].probs, probs[:, -1])


if __name__ == '__main__':
  absltest.main()
